/* JB FZGC2

 Jan Bochmann 2/2007

 Stand/Last changes:  2009-02-08

DE: Compiler zum Erstellen von Archivdateien aus einzelnen Dateien
    sowie umgekehrt zum Erstellen selbiger aus dem jeweiligen Archiv

EN: Compiler for creating archive files from single files
    and back for creating them from the respective archive files

    ".fz2" => "bahn.fz2"
    ".fz4" => "bahn.fz4"
    ".gz1" => "bahn.fw1", "bahn.fx1", "name.uz1"
    ".gz2" => "bahn.fw2", "bahn.fx2", "name.uz2"
    ".gz4" => "bahn.fw4", "bahn.fx4", "name.uz4"

DE: Aufrufparameter: Werden bei Start ohne Parameter angezeigt
    Leerzeichen ist Trennzeichen, d.h. Namen, die Leerzeichen enthalten,
    werden nicht unterstuetzt.

EN: Arguments for call: Are displayed when starting without parameters
    Empty space is separator, i.e. names containing spaces are not supported.
   */

/* ------------------------------------------------------------------------ */

/* DE: Dies ist eigtl. ein Konsolenprogramm, aber folgende Windows-Funktionen
       werden verwendet:
   EN: Program is made for console/command line,
       however the following MS-Windows functions are used:

       MessageBox()
       GetUserDefaultLangID()
     */

#include <windows.h>

#pragma hdrstop

/* ------------------------------------------------------------------------
DE: Format der Ausgabedateien:
    (Aufbau der Einzeldateien siehe extra Dokumentation)

  Zoom2 Fahrzeuge:

  1. Vorspann bis CTRL Z
  2. BYTE IDENT_CODE[4]={'F','Z','X','2'}
  3. Versionsnummer INT32 = {0x38400000} LOWORD-HIWORD je LOBYTE-HIBYTE
  3. Vorspann fr schnelleres Suchen mit lseek()
      INT32  Fzg.nummer ( 0 == Ende, danach kein INT32 )
      INT32  Dateioffset fr lseek() ab Position hinter dem INT32
             zum Erreichen des Datenblocks
  4. Fahrzeuge (Datenbloecke):
      INT32    Fzg.nummer (dieselbe wie im Vorspann 3.)
      INT32    Laenge der Einzeldatei in BYTE
      INT32    Datum/Zeit der Datei
      BYTE[]   komplette .fz2-Datei (binaer kopiert)
  5. keine Endekennung

  Zoom4 Fahrzeuge:

  1. Vorspann bis CTRL Z
  2. BYTE IDENT_CODE[4]={'F','Z','X','4'}
  3. Versionsnummer INT32 = {0x38500000} LOWORD-HIWORD je LOBYTE-HIBYTE
  3. Vorspann fr schnelleres Suchen mit lseek()
      INT32  Fzg.nummer ( 0 == Ende, danach kein INT32 )
      INT32  Dateioffset fr lseek() ab Position hinter dem INT32
             zum Erreichen des Datenblocks
  4. Fahrzeuge (Datenbloecke):
      INT32    Fzg.nummer (dieselbe wie im Vorspann 3.)
      INT32    Laenge der Einzeldatei in BYTE
      INT32    Datum/Zeit der Datei
      BYTE[]   komplette .fz4-Datei (binaer kopiert)
  5. keine Endekennung

  Zoom1/2/4 Landschaft, nutzerdefiniert:

  1. Vorspann bis CTRL Z
  2. BYTE IDENT_CODE[3] ={'U','Z','X'}
  3. BYTE ZoomFaktorID = '1' oder '2' oder '4'
  4. Vorspann fr schnelleres Suchen mit lseek()
      INT32  Elementnummer lokal in der Datei 0..89
                 (-1 == Endekennzeichen, danach keine weiteren Daten)
      INT32  Bauform = 0
      INT32  FwSig = 0
      INT32  Phase = Animationsphase
      INT32  Alt   = Alternative (0 oder 1..4 fuer "a".."d")
      INT32  Dateioffset fr lseek() ab Position hinter der Elementnummer
             zum Erreichen des Datenblocks
  5. Elemente (Datenbloecke):
      INT32  Elementnummer (die selbe wie im Vorspann)
      INT32  Bauform (die selbe wie im Vorspann)
      INT32  FwSig (die selbe wie im Vorspann)
      INT32  Phase (die selbe wie im Vorspann)
      INT32  Alt   (die selbe wie im Vorspann)
      INT32  Laenge der Einzeldatei in BYTE
      INT32  Datum/Zeit der Datei
      BYTE[] komplette .gz1/2/4-Datei (binaer kopiert)
  6. keine Endekennung

  Zoom1/2/4 Standard (Landschaft und Fahrwegdaten):

  1. Vorspann bis CTRL Z
  2. BYTE IDENT_CODE[3] ={'G','Z','X'}
  3. BYTE ZoomFaktorID = '1' oder '2' oder '4'
  4. Vorspann fr schnelleres Suchen mit lseek()
      INT32  Elementnummer (global in BAHN)
                (-1 == Endekennzeichen, danach keine weiteren Daten)
      INT32  Bauform (bei Landschaft immer = 0)
      INT32  FwSig: bei Landschaft = 0
                    bei Fahrweg: HIWORD = Fahrweg 1..4 (FAHRWEGTYP_xxx+1)
                                 LOWORD = bei Signalen Signalphase 1 oder 2
                                          sonst = 0
      INT32  Phase = Animationsphase (bei Fahrweg immer = 0)
      INT32  Alt   = Alternative (0 oder 1..4 fuer "a".."d")
                     bei Fahrweg immer = 0
      INT32  Dateioffset fr lseek() ab Position hinter der Elementnummer
             zum Erreichen des Datenblocks
  5. Elemente (Datenbloecke):
      INT32  Elementnummer (die selbe wie im Vorspann)
      INT32  Bauform (die selbe wie im Vorspann)
      INT32  FwSig (die selbe wie im Vorspann)
      INT32  Phase (die selbe wie im Vorspann)
      INT32  Alt   (die selbe wie im Vorspann)
      INT32  Laenge der Einzeldatei in BYTE
      INT32  Datum/Zeit der Datei
      BYTE[] komplette .gz1/2/4-Datei (binaer kopiert)
  6. keine Endekennung

EN: Output file format:
    (for structure of single files see extra documentation)

  Zoom2 Vehicles:

  1. Header (text) until CTRL Z
  2. BYTE IDENT_CODE[4]={'F','Z','X','2'}
  3. Version identifier INT32 = {0x38400000} LOWORD-HIWORD each LOBYTE-HIBYTE
  3. Header for faster search using lseek()
      INT32  vehicle number ( 0 == terminator, after no INT32 following)
      INT32  file offset for lseek() from position behind the INT32
             for reaching the data block
  4. Vehicles (data blocks):
      INT32    vehicle number (same as in header 3.)
      INT32    length of single file in BYTE
      INT32    date/time of single file
      BYTE[]   complete .fz2 file (binary copied)
  5. no terminator

  Zoom4 Vehicles:

  1. Header (text) until CTRL Z
  2. BYTE IDENT_CODE[4]={'F','Z','X','4'}
  3. Version identifier INT32 = {0x38500000} LOWORD-HIWORD each LOBYTE-HIBYTE
  3. Header for faster search using lseek()
      INT32  vehicle number ( 0 == terminator, after no INT32 following)
      INT32  file offset for lseek() from position behind the INT32
             for reaching the data block
  4. Vehicles (data blocks):
      INT32    vehicle number (same as in header 3.)
      INT32    length of single file in BYTE
      INT32    date/time of single file
      BYTE[]   complete .fz4 file (binary copied)
  5. no terminator

  Zoom1/2/4 Scenery, user-defined:

  1. Header (text) until CTRL Z
  2. BYTE IDENT_CODE[3] ={'U','Z','X'}
  3. BYTE ZoomFaktorID = '1' or '2' or '4'
  4. Header for faster search using lseek()
      INT32  Element number locally in the file 0..89
                 (-1 == terminator, after it no data)
      INT32  Bauform = 0
      INT32  FwSig = 0
      INT32  Phase = Animation phase
      INT32  Alt   = Alternative/Variant (0 or 1..4 for "a".."d")
      INT32  file offset for lseek() from position behind the element number
             for reaching the data block
  5. Elements (data blocks):
      INT32  Element number (same as in header)
      INT32  Bauform (same as in header)
      INT32  FwSig (same as in header)
      INT32  Phase (same as in header)
      INT32  Alt   (same as in header)
      INT32  length of single file in BYTE
      INT32  date/time of single file
      BYTE[] complete .gz1/2/4 file (binary copied)
  6. no terminator

  Zoom1/2/4 Standard (scenery and driving way add-on):

  1. Header (text) until CTRL Z
  2. BYTE IDENT_CODE[3] ={'G','Z','X'}
  3. BYTE ZoomFaktorID = '1' or '2' or '4'
  4. Header for faster search using lseek()
      INT32  Element number (globally in BAHN)
                (-1 == terminator, after it no data)
      INT32  Bauform: for scenery always = 0
                      for driving way: graphics version
      INT32  FwSig: for scenery = 0
                    for driv.way: HIWORD = dr.way 1..4 (FAHRWEGTYP_xxx+1)
                                  LOWORD = for signals signal phase 1 or 2
                                           otherwise = 0
      INT32  Phase = Animation phase (for driv.way always = 0)
      INT32  Alt   = Alternative/Variant (0 or 1..4 for "a".."d")
                     for driving way = 0
      INT32  file offset for lseek() from position behind the element number
             for reaching the data block
  5. Elements (data blocks):
      INT32  Element number (same as in header)
      INT32  Bauform (same as in header)
      INT32  FwSig (same as in header)
      INT32  Phase (same as in header)
      INT32  Alt   (same as in header)
      INT32  length of single file in BYTE
      INT32  date/time of single file
      BYTE[] complete .gz1/2/4 file (binary copied)
  6. no terminator

 */

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <dir.h>
#include <string.h>
#include <fcntl.h>
#include <alloc.h>

#define FASTCALL        _fastcall
#define CONST           const
#define STATIC          static
/* typedef void VOID; */
typedef unsigned char   BYTE;
typedef   signed char   CHAR8;
typedef   signed short  INT16;
typedef   signed short  CHAR16;
typedef unsigned short  WORD16;
typedef   signed int    INT32;
typedef unsigned int    WORD32;
typedef int             BOOL;
typedef VOID *          PVOID;
typedef VOID CONST *    PCVOID;
typedef BYTE *          PBYTE;
typedef CHAR8 *         PCHAR8;
typedef CHAR8 CONST *   PCCHAR8;
typedef CHAR16 *        PCHAR16;
typedef CHAR16 CONST *  PCCHAR16;
typedef WORD16 *        PWORD16;
typedef INT16 *         PINT16;
typedef WORD32 *        PWORD32;
typedef INT32 *         PINT32;
typedef BOOL *          PBOOL;
#define TRUE  1
#define FALSE 0

/* #define LOWORD(l)	    ((WORD16)(l))
#define HIWORD(l)	    ((WORD16) ((l) >> 16))
#define LOBYTE(w)	    ((w) & 0xFF)
#define HIBYTE(w)	    ((w) >> 8) */

#define CTRL_Z  26

/* Testuntersttzung */
#undef TEST10
#undef TEST20
#undef TEST201
#undef TEST310

/* Sprachen / Languages (not all supported) */
#define SPRACHE_MIN         0
#define SPRACHE_ENG         0
#define SPRACHE_DEU         1
#define SPRACHE_POL         2
#define SPRACHE_ESP         3
#define SPRACHE_CZE         4
#define SPRACHE_MAX         4
#define SPRACHE_ANZAHL      (SPRACHE_MAX-SPRACHE_MIN+1)
/* #define SPRACHE_FRA         2 */
/* #define SPRACHE_NLD         3 */

/* Global variables */
CONST CHAR8 IDENT_CODE_FZX[3] = { 'F', 'Z', 'X' };
CONST CHAR8 IDENT_CODE_GZX[3] = { 'G', 'Z', 'X' };
CONST CHAR8 IDENT_CODE_UZX[3] = { 'U', 'Z', 'X' };

/* File extensions single files */
#define FEXT_ZOOM2FZG  "fz2"  /* b384,b385beta1..3: "fzz" */
#define FEXT_ZOOM4FZG  "fz4"
#define FEXT_ZOOM1GZG  "gz1"
#define FEXT_ZOOM2GZG  "gz2"
#define FEXT_ZOOM4GZG  "gz4"
/* File extensions archive files (+ '0'+zoomfaktor) */
#define FEXT_ARCZOOMFZG  ".fz"
#define FEXT_ARCZOOMGZG  ".gz"
#define FEXT_ARCZOOMUZG  ".uz"

/* Commands and options are case insensitive */

#define COMMAND_EXTRACT    'X'
#define COMMAND_COMBINEFZG 'V'
#define COMMAND_COMBINEGZG 'S'
#define COMMAND_COMBINEUZG 'U'
#define COMMAND_ZOOM1      '1'
#define COMMAND_ZOOM2      '2'
#define COMMAND_ZOOM4      '4'

#define OPTION_OVERWRITE   'O'
#define OPTION_SIMULATE    'S'

 /* Maximum count of vehicles in one vehicle set */
#define FZGF_MAXFZGZAHL      16

typedef struct FZGELEM * PFZGELEM;
typedef struct FZGELEM {
                  PFZGELEM pNext; /* Pointer to next, =NULL if last one */
                  INT32    FzgNr; /* Number ((Satznummer << 4) | LokaleNr ) */
		  INT32    Len;   /* Length of file in BYTE */
                  INT32    Zeit;  /* Date+Time, see there */
                  CHAR8    PfName[MAXPATH]; /* Path+Name (not very efficient...) */
	       } FZGELEM;

typedef struct GZGELEM * PGZGELEM;
typedef struct GZGELEM {
                  PGZGELEM pNext; /* Pointer to next, =NULL if last one */
                  INT32    SymNr; /* Number, Standard: global
                                             User-defined: local in the file */
                  INT32    Bauform; /* Bauform/Version
                                       (for std. driving way, else = 0) */
                  INT32    FwSig;   /* HIWORD: Fahrweg/Driving way 1..4
                                       (for std. driving way, else = 0)
                                       FAHRWEGTYP_xxx + 1
                                       LOWORD: Signal Phase = 1 or 2
                                       (for blinking signals
                                        i.e. grade crossings
                                        else (mostly) = 0) */
                  INT32    Phase;   /* Animation Phase
                                       (for driving way = 0) */
                  INT32    Alt;     /* Alternative (0 or 1..4) */
		  INT32    Len;   /* Length of file in BYTE */
                  INT32    Zeit;  /* Date+Time, see there */
                  CHAR8    PfName[MAXPATH]; /* Path+Name (not very efficient...) */
	       } GZGELEM;

/* Count of stored file informations */
STATIC INT32  Anzahl_Stored = 0;

STATIC INT32  Sprach_Code;

typedef struct TXTELEM * PTXTELEM;
typedef struct TXTELEM CONST * PCTXTELEM;
typedef struct TXTELEM {
                  CONST INT32 TxtNr; /* Number of text */
                  /* Pointer to Text: SPRACHE_MIN.._MAX
                                      _ENG,_DEU,_POL,_ESP,_CZE */
                  CONST PCCHAR8 pTxt[ SPRACHE_ANZAHL ];
	       } TXTELEM;

#define TXTNR_UNDEF  -1
#define TXTNR_000     0
#define TXTNR_001     1
#define TXTNR_002     2
#define TXTNR_003     3
#define TXTNR_004     4
#define TXTNR_005     5
#define TXTNR_006     6
#define TXTNR_007     7
#define TXTNR_008     8
#define TXTNR_009     9
#define TXTNR_010    10
#define TXTNR_011    11
#define TXTNR_012    12
#define TXTNR_013    13
#define TXTNR_014    14
#define TXTNR_015    15
#define TXTNR_016    16
#define TXTNR_017    17
#define TXTNR_F_LESEN  18
#define TXTNR_F_DATEN  19

#define MIN_FAHRWEG             0x0000
#define MAX_FAHRWEG             0x3FFF
#define USERGSYMZAHL_PER_DATEI  90  /* bis B384: (2*36) */
/* kleinstes+groesstes Landschaftssymbol
   inkl. nutzerdef. */
#define MIN_LSYMBOL      0xC000
#define MAX_LSYMBOL      0xFFFF
/* Animationsphasen von Landschaftssymbolen */
#define MIN_ANIPHASE     0
#define MAX_ANIPHASE    99
/* Alternativ-Varianten von Landschaftssymbolen */
#define KEINE_ALTERNATIVE 0  /* = alle Alternativen wenn mehrere ex. */
#define MIN_ALTERNATIVE   1
#define MAX_ALTERNATIVE   4
#define MAX_ALTERNATIVZAHL ( MAX_ALTERNATIVE - MIN_ALTERNATIVE + 1 )

/* Fahrwegtypen / Driving way types */
#define ANZAHL_FAHRWEGTYPEN  4
#define FAHRWEGTYP_GLSTR     0
#define FAHRWEGTYP_GLBBK     1
#define FAHRWEGTYP_H2O       2
#define FAHRWEGTYP_BUS       3
CONST CHAR Fahrwegzeichen[ ANZAHL_FAHRWEGTYPEN ] = {'s','t','h','b'};
#define ZEICHEN_NUTZERDEF 'u'

#define FILEBUFSIZE 4096
STATIC BYTE   WriteBuf[ FILEBUFSIZE ]; /* Dateipuffer (Schreiben) */
STATIC INT32  WritePos = 0;      /* aktuelle Position beim Schreiben */
STATIC BYTE   ReadBuf[ FILEBUFSIZE ]; /* Dateipuffer (Lesen) */
STATIC INT32  ReadPos = 0;      /* aktuelle Position beim Lesen */

STATIC PFZGELEM p_FzgElemente = NULL;
STATIC PGZGELEM p_GzgElemente = NULL;

/* lokale Fktn. */

STATIC VOID FASTCALL InfoUserEnd( VOID );
STATIC VOID FASTCALL InitAll( VOID );
STATIC VOID FASTCALL QuitAll( VOID );
STATIC INT32 FASTCALL EinpackenFzg( PCHAR8 FileName, PCHAR8 Directory,
                                    INT32 Zoomfaktor, BOOL Simulieren );
STATIC INT32 FASTCALL EinpackenGzg( PCHAR8 FileName, PCHAR8 Directory,
                                    INT32 Zoomfaktor, BOOL Nutzerdef,
                                    BOOL Simulieren );
STATIC INT32 FASTCALL Auspacken( PCHAR8 FileName, PCHAR8 Directory,
                                 BOOL Ueberschreiben, BOOL Simulieren,
                                 PINT32 pFound,
                                 PINT32 pCopied, PINT32 pSkipped );
STATIC INT32 FASTCALL WriteSeekTableFzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL WriteDatenFzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL StoreNameFzg( PCHAR8 pPfad, PCHAR8 pName,
                                    INT32 len, INT32 datum, INT32 zeit );
STATIC INT32 FASTCALL WriteSeekTableGzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL WriteDatenGzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL StoreNameGzg( PCHAR8 pPfad, PCHAR8 pName,
                                    INT32 Len, INT32 datum, INT32 zeit,
                                    BOOL Nutzerdef, PCHAR Archivname );
STATIC VOID FASTCALL Fehler( PCCHAR8 pText );
STATIC VOID FASTCALL Trace( PCCHAR8 pText );
STATIC VOID FASTCALL Meldung( PCCHAR8 pText, PCCHAR8 pTitel );
STATIC CHAR FASTCALL Caps( CHAR zei );
STATIC VOID FASTCALL CapsString( PCHAR8 Text );
STATIC BOOL FASTCALL DateiGibtEs( PCCHAR8 pName );
STATIC FILE * FASTCALL OpenWrite( PCCHAR8 pName );
STATIC INT32 FASTCALL WriteByte( FILE * file2, BYTE b );
STATIC INT32 FASTCALL WriteWord16( FILE * file2, WORD16 w );
STATIC INT32 FASTCALL WriteInt32( FILE * file2, INT32 i );
STATIC INT32 FASTCALL CloseWrite( FILE * file2 );
STATIC FILE * FASTCALL OpenRead( PCCHAR8 pName );
STATIC INT32 FASTCALL ReadByte( FILE * file, PBYTE pB );
STATIC INT32 FASTCALL ReadWord16( FILE * file, PWORD16 pW );
STATIC INT32 FASTCALL ReadInt32( FILE * file, PINT32 pI );
STATIC INT32 FASTCALL CloseRead( FILE * file );
STATIC PVOID FASTCALL AllocateMemory( INT32 len );
STATIC VOID FASTCALL FreeMemory( PVOID pDat );
STATIC INT32 FASTCALL ReadOverSpacePos( PCCHAR8 pBuf );
STATIC INT32 FASTCALL DelSpaces( PCHAR8 pBuf );
STATIC INT32 FASTCALL SetDateiAttr( PCCHAR8 pPfad, INT32 Zeit );
STATIC INT32 FASTCALL GetSprache( VOID );
STATIC PCCHAR8 FASTCALL GetText( INT32 Textnummer );
STATIC VOID FASTCALL MemMove1( PVOID pDest, PCVOID pSrc, INT32 len );
#define MemCopy1( pD, pS, nL ) MemMove1( pD, pS, nL )
STATIC INT32 FASTCALL GetGzgDateiName( INT32 GSym, INT32 Bauform, INT32 FwSig,
                                       INT32 AddInfo, INT32 Phase,
                                       PCHAR8 pFName, INT32 Zf,
                                       PCCHAR8 pNutzerName );
STATIC VOID FASTCALL strcadd( PCHAR8 pStr, CHAR8 Zei );
STATIC VOID FASTCALL SubExt( PCHAR8 pName );

/*--------------------------------------------------------------*/
/* MAIN */

#pragma argsused
INT32 main( INT32 argc, PCHAR8 argv[] )
  {
   CHAR8  ArcFileName[ MAXPATH ];
   CHAR8  DirectoryName[ MAXPATH ];
   FILE   * file1;
   INT32  rc, optbuflen, optzei0, optzei1, zoomfaktor, len;
   INT32  anzfound, anzcopy, anzskip;
   CHAR8  errbuf[MAXPATH+80], optbuf[10];
   PCHAR8 pOptBuf;
   BOOL   auspacken, fahrzeuge, nutzerdef, overwrite, simulate;

     /* Sprache ermitteln / Get Language */
     Sprach_Code = GetSprache();

     printf( "\nJB Data-Combine %s 0.009 02/2009\n", GetText( TXTNR_001 ) );

     if ( _argc < 4 )
        /* 0 is the exe file self, 1..4 are parameters */
       {
        InfoUserEnd();
        return 1;
       }

     optbuflen = strlen( _argv[ 1 ] ) + 1;
     MemCopy1( optbuf, _argv[ 1 ], min( optbuflen, sizeof( optbuf ) ) );
     optbuf[ sizeof( optbuf ) - 2 ] = '\0';

     DelSpaces( optbuf );
     pOptBuf = optbuf;

     if (( *pOptBuf == '-' ) || ( *pOptBuf == '/' ))
       {
        pOptBuf++;
        DelSpaces( pOptBuf );
       }

     optzei0 = Caps( *pOptBuf++ );
     if ( optzei0 == COMMAND_EXTRACT )
        auspacken = TRUE;
     else
       {
        auspacken = fahrzeuge = nutzerdef = FALSE;
        if ( optzei0 == COMMAND_COMBINEFZG )
           fahrzeuge = TRUE;
        else
          {
           if ( optzei0 == COMMAND_COMBINEUZG )
              nutzerdef = TRUE;
           else
             {
              if ( optzei0 != COMMAND_COMBINEGZG )
                {
                 InfoUserEnd();
                 return 1;
                }
             }
          }
       }

     if ( !auspacken )
       {
        DelSpaces( pOptBuf );
        optzei1 = Caps( *pOptBuf++ );
        if ( optzei1 == COMMAND_ZOOM1 )
           zoomfaktor = 1;
        else
          {
           if ( optzei1 == COMMAND_ZOOM2 )
              zoomfaktor = 2;
           else
             {
              if ( optzei1 == COMMAND_ZOOM4 )
                 zoomfaktor = 4;
              else
                {
                 InfoUserEnd();
                 return 1;
                }
             }
          }
       }
     else /* Auspacken */
        pOptBuf++;

     /* Argument 2: Directory */
     strcpy( DirectoryName, _argv[ 2 ] );
     len = strlen( DirectoryName );
     if ( len > 0 )
       {
        if ( DirectoryName[ len - 1 ] == '\\' )
           DirectoryName[ len - 1 ] = '\0';
       }
       
     /* Argument 3: Filename */
     strcpy( ArcFileName, _argv[ 3 ] );
     if ( !auspacken )
       {
        if ( strchr( ArcFileName, '.' ) == NULL )
          {
           if ( fahrzeuge )
             {
              strcat( ArcFileName, FEXT_ARCZOOMFZG );
             }
           else
             {
              if ( nutzerdef )
                {
                 strcat( ArcFileName, FEXT_ARCZOOMUZG );
                }
              else
                {
                 strcat( ArcFileName, FEXT_ARCZOOMGZG );
                }
             }
           strcadd( ArcFileName, LOBYTE( '0'+zoomfaktor ) );
          }
       }   

     /* Argument 4: weitere Optionen / More options */
     overwrite = simulate = FALSE;
     if ( _argc >= 5 )
       {
        optbuflen = strlen( _argv[ 4 ] ) + 1;
        MemCopy1( optbuf, _argv[ 4 ], min( optbuflen, sizeof( optbuf ) ) );
        optbuf[ sizeof( optbuf ) - 2 ] = '\0';

        DelSpaces( optbuf );
        pOptBuf = optbuf;

        if (( *pOptBuf == '-' ) || ( *pOptBuf == '/' ))
          {
           pOptBuf++;
           DelSpaces( pOptBuf );
          }

        while ( *pOptBuf != '\0' )
          {
           optzei0 = Caps( *pOptBuf++ );
           if ( optzei0 == OPTION_OVERWRITE )
              overwrite = TRUE;
           else
              if ( optzei0 == OPTION_SIMULATE )
                 simulate = TRUE;
          }       
       }

     file1 = OpenRead( ArcFileName );
     CloseRead( file1 );

     if (file1 == NULL)
       {
        if ( auspacken )
          {
           sprintf( errbuf, GetText( TXTNR_002 ), ArcFileName );
           Fehler( errbuf );
           exit(1);
          }
       }
     else /* opened OK */
       {
        if ( !auspacken )
          {
           if ( !overwrite )
             {
              sprintf( errbuf, GetText( TXTNR_003 ), ArcFileName );
              Fehler( errbuf );
              exit(1);
             } 
          }
       }

     InitAll();

     if ( auspacken )
       {
        rc = Auspacken( ArcFileName, DirectoryName, overwrite, simulate,
                        &anzfound, &anzcopy, &anzskip );

        if ( rc == 0 )
          {
           sprintf( errbuf, GetText( TXTNR_007 ), anzfound, anzcopy, anzskip );
           Meldung( errbuf, GetText( TXTNR_008 ) );
          }
        else
          {
           sprintf( errbuf, "%s%i",
                            GetText( TXTNR_009 ),
                            rc );
           Fehler( errbuf );
          }
       }
     else /* Einpacken */
       {
        if ( fahrzeuge )
           rc = EinpackenFzg( ArcFileName, DirectoryName, zoomfaktor, simulate );
        else
           rc = EinpackenGzg( ArcFileName, DirectoryName, zoomfaktor,
                              nutzerdef, simulate );

        switch ( rc )
          {
           case 0: /* OK */
              sprintf( errbuf, GetText( TXTNR_010 ),
                       Anzahl_Stored );
              Meldung( errbuf, GetText( TXTNR_008 ) );
              break;

           case 1: /* No file found */
              Fehler( GetText( TXTNR_014 ) );
              break;

           case 2:  /* out of mem */
           case 12:
              Fehler( GetText( TXTNR_015 ) );
              break;

           case 3: /* No archive created */
              Fehler( GetText( TXTNR_016 ) );
              break;

           case 4: /* Write error */
              Fehler( GetText( TXTNR_017 ) );
              break;

           case 11: /* Open/Read error */
              Fehler( GetText( TXTNR_F_LESEN ) );
              break;

           case 13: /* Data error */
              Fehler( GetText( TXTNR_F_DATEN ) );
              break;

           default: /* Something unknown: telling the rc */
              sprintf( errbuf, "%s%i",
                                GetText( TXTNR_011 ), rc );
              Fehler( errbuf );
          }
       }

     return 0;
    }

/* --------------------------------------------------------------------- */

VOID FASTCALL InitAll( VOID )
    /* Initialisierungen */
    {
     p_FzgElemente = NULL;
     p_GzgElemente = NULL;
     Anzahl_Stored = 0;
    }

/* --------------------------------------------------------------------- */

VOID FASTCALL QuitAll( VOID )
    /* Freigaben */
  {
   PFZGELEM pT1, pT2;
   PGZGELEM pG1, pG2;

     pT1 = p_FzgElemente;
     p_FzgElemente = NULL;
     while ( pT1 != NULL )
       {
        pT2 = pT1->pNext;
        FreeMemory( pT1 );
        pT1 = pT2;
       }

     pG1 = p_GzgElemente;
     p_GzgElemente = NULL;
     while ( pG1 != NULL )
       {
        pG2 = pG1->pNext;
        FreeMemory( pG1 );
        pG1 = pG2;
       }
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL EinpackenFzg( PCHAR FileName, PCHAR Directory,
                                    INT32 Zoomfaktor, BOOL Simulieren )
    /* Packt ein (Fahrzeuge)
       IN:  Filename = Archivdatei (wird ggf. ueberschrieben)
                       kann Pfad enthalten
            Directory = Quellverzeichnis           
            ZoomFaktor = 2 oder 4
            Simulieren = TRUE: nichts wirklich schreiben
       OUT: 0 = OK
            1 = keine einzige Datei gefunden
            2 = Speicher reicht nicht
            3 = Archivdatei nicht erzeugt
            4 = Schreibfehler Archivdatei
            5 = falscher Parameter */
  {
   CHAR8 Pfad[ MAXPATH ], PfadMask[ MAXPATH ];
   struct ffblk fblk;
   FILE * file2;
   INT32  rc, i;
   CONST CHAR8 Vsp[] = "JB BAHN Vehicle gfx Zoom";
   PCCHAR8 pZei;

     if (( Zoomfaktor != 2 ) && ( Zoomfaktor != 4 ))
        return 5;

     if ( strlen( FileName ) + 60 >= sizeof( Pfad ) )
        return 5;

     strcpy( Pfad, Directory );

     strcpy( PfadMask, Pfad );
     strcat( PfadMask, "\\*." );
     if ( Zoomfaktor == 2 )
        strcat( PfadMask, FEXT_ZOOM2FZG );
     else
        strcat( PfadMask, FEXT_ZOOM4FZG );

     if ( findfirst( PfadMask, &fblk, 0 ) != 0 )
        return 1;

     if ( StoreNameFzg( Pfad, fblk.ff_name,
                        fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime ) )
        return 2;

     while ( findnext( &fblk ) == 0 )
       {
        if ( StoreNameFzg( Pfad, fblk.ff_name,
                           fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime ) )
           return 2;
       }

     if ( ! Simulieren )
       {
        file2 = OpenWrite( FileName );
        if ( file2 == NULL )
           return 3;

        /* Vorspann schreiben */
        pZei = Vsp;
        while ( *pZei != '\0' )
          {
           rc = WriteByte( file2, *pZei++ );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        rc = WriteByte( file2, CTRL_Z );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        /* Kennung schreiben */
        for ( i = 0; i < sizeof( IDENT_CODE_FZX ); i++ )
          {
           rc = WriteByte( file2, IDENT_CODE_FZX[i] );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
       }   

     /* Seektabelle schreiben */
     rc = WriteSeekTableFzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     /* einzelne Dateien kopieren */
     rc = WriteDatenFzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     if ( ! Simulieren )
       {
        if ( CloseWrite( file2 ) )
           return 4;
       }    

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL EinpackenGzg( PCHAR FileName, PCHAR Directory,
                                    INT32 Zoomfaktor, BOOL Nutzerdef,
                                    BOOL Simulieren )
    /* Packt ein (Landschaft)
       IN:  Filename = Archivdatei (wird ggf. ueberschrieben)
                       kann Pfad enthalten
            Directory = Quellverzeichnis
            ZoomFaktor = 1, 2 oder 4
            Nutzerdef = TRUE: nutzerdefinierte
            Simulieren = TRUE: nichts wirklich schreiben
       OUT: 0 = OK
            1 = keine einzige Datei gefunden
            3 = Archivdatei nicht erzeugt
            4 = Schreibfehler Archivdatei
            5 = falscher Parameter
           11 = Problem beim Oeffnen/Datenlesen
           12 = Speicher reicht nicht
           13 = Datenfehler */
  {
   CHAR8 Pfad[ MAXPATH ], PfadMask[ MAXPATH ], Dateiname[ MAXPATH ];
   struct ffblk fblk;
   FILE * file2;
   INT32  rc, i;
   STATIC CHAR Vsp1[] = "JB BAHN Scenery gfx Zoom";
   STATIC CHAR Vsp2[] = "\r\ncreated by FZGC2";
   PCCHAR8 pZei;
   PCCHAR8 pIDC;

     if (( Zoomfaktor != 1 ) && ( Zoomfaktor != 2 ) && ( Zoomfaktor != 4 ))
        return 5;

     if ( strlen( FileName ) + 60 >= sizeof( Pfad ) )
        return 5;

     strcpy( Pfad, Directory );

     if ( Nutzerdef )
       {
        sprintf( PfadMask, "%s\\%c*.", Pfad, ZEICHEN_NUTZERDEF );
        rc = fnsplit( FileName, NULL, NULL, Dateiname, NULL );
        if (( rc & WILDCARDS ) || !( rc & FILENAME ))
           return 5;
       }
     else
       {
        strcpy( PfadMask, Pfad );
        strcat( PfadMask, "\\*." );
       }
     if ( Zoomfaktor == 2 )
        strcat( PfadMask, FEXT_ZOOM2GZG );
     else
        if ( Zoomfaktor == 4 )
           strcat( PfadMask, FEXT_ZOOM4GZG );
        else
           strcat( PfadMask, FEXT_ZOOM1GZG );

     /* Dateien erfassen */

     if ( findfirst( PfadMask, &fblk, 0 ) != 0 )
       {
        Fehler( GetText( TXTNR_014 ) );
        return 1;
       }

     rc = StoreNameGzg( Pfad, fblk.ff_name,
                        fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime,
                        Nutzerdef, Dateiname );
     if ( rc )
        return 10+rc;

     while ( findnext( &fblk ) == 0 )
       {
        rc = StoreNameGzg( Pfad, fblk.ff_name,
                           fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime,
                           Nutzerdef, Dateiname );
        if ( rc )
           return 10+rc;
       }

     /* eigtl. schreiben */

     if ( ! Simulieren )
       {
        file2 = OpenWrite( FileName );
        if ( file2 == NULL )
           return 3;

        /* Vorspann schreiben */
        pZei = Vsp1;
        while ( *pZei != '\0' )
          {
           rc = WriteByte( file2, *pZei++ );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
        pZei = Vsp2;
        while ( *pZei != '\0' )
          {
           rc = WriteByte( file2, *pZei++ );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }

        rc = WriteByte( file2, CTRL_Z );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        /* Kennung schreiben */
        if ( Nutzerdef )
           pIDC = IDENT_CODE_UZX;
        else
           pIDC = IDENT_CODE_GZX;
        for ( i = 0; i < sizeof( IDENT_CODE_GZX ); i++ )
          {
           rc = WriteByte( file2, pIDC[i] );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
       }

     /* Seektabelle schreiben */
     rc = WriteSeekTableGzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     /* einzelne Dateien kopieren */
     rc = WriteDatenGzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     if ( ! Simulieren )
       {
        if ( CloseWrite( file2 ) )
           return 4;
       }

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL StoreNameFzg( PCHAR pPfad, PCHAR pName,
                                    INT32 len, INT32 datum, INT32 zeit )
/* speichert Name der Einzeldatei und notwendige Eigenschaften
   (Fahrzeuge)
   IN: pPfad = Pfad
       pName = Dateiname
       len   = Laenge in BYTE
       datum = Bit0..4 Tag, Bit5..8 Monat, Bit9..15 Jahr seit 1980
       zeit  = Bit0..4 sec/2, Bit5..10 min, Bit11..15 h
  OUT: 0 OK
       1 Problem beim Oeffnen/Datenlesen
       2 Speicher reicht nicht
       3 Datenfehler */
  {
   INT32    satznr, locfzgnr, fzgnr;
   PFZGELEM pEl, pV, pN;

     if ( len <= 0 )
        return 1;

     if ( sscanf( pName, "%u-%u", &satznr, &locfzgnr ) != 2 )
        return 3;

     if (( satznr < 0 ) || ( locfzgnr < 0 ) || ( locfzgnr > FZGF_MAXFZGZAHL ))
        return 3;

     fzgnr = ( satznr << 4 ) | locfzgnr;

     pEl = AllocateMemory( sizeof( FZGELEM ) );
     if ( pEl == NULL )
        return 2;

     pEl->FzgNr = fzgnr;
     pEl->Len   = len;
     pEl->Zeit  = (datum << 16) | zeit;
     strcpy( pEl->PfName, pPfad );
     strcat( pEl->PfName, "\\" );
     strcat( pEl->PfName, pName );

     /* InQueue/Einketten, sortiert */
     pV = NULL;
     pN = p_FzgElemente;

     while ( pN != NULL )
       {
        if ( pN->FzgNr > fzgnr )
           break;
        pV = pN;
        pN = pN->pNext;
       }

     if ( pV == NULL )
       {
        pEl->pNext = p_FzgElemente;
        p_FzgElemente = pEl;
       }
     else
       {
        pEl->pNext = pV->pNext;
        pV->pNext = pEl;
       }

     Anzahl_Stored++;

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteSeekTableFzg( FILE * file2, BOOL Simulieren )
 /* Schreibt Tabelle der FzgNummern und seek()-Positionen
    (Fahrzeuge)
    IN:  file2 = Handle Datei
         Simulieren = TRUE: nichts wirklich schreiben
   OUT:  0 = OK
         1 = Schreibfehler (Datei geschlossen) */
  {
   INT32    len;
   PFZGELEM pTab = p_FzgElemente;

     /* Laenge der Seek-Tabelle selbst inkl. Endekz. */
     len = Anzahl_Stored * (sizeof(INT32) + sizeof(INT32)) + sizeof(INT32);

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->FzgNr ))
              return 1;
          }

        /* abzgl. des eigenen Seek-Eintrags, der wurde ja dann schon gelesen */
        len -= (sizeof(INT32) + sizeof(INT32));

        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, len ) )
              return 1;
          }

        /* Gesamtlaenge, + vor der Datei noch die Nr., eigene Laenge + Zeit
           (ist alles etwas redundant aber dadurch einfacher zu handhaben) */
        len += pTab->Len + 3 * sizeof( INT32 );

        pTab = pTab->pNext;
       }

     /* Endekennzeichen */
     if ( ! Simulieren )
       {
        if ( WriteInt32( file2, 0 ) )
           return 1;
       }    

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteDatenFzg( FILE * file2, BOOL Simulieren )
 /* Schreibt eigentlichen Datenteil, d.h. kopiert Dateien
    (Fahrzeuge)
    IN:  file2 = Handle Datei
         Simulieren = TRUE: nichts wirklich schreiben
   OUT:  0 = OK
         1 = Schreibfehler
         2 = Lesefehler
    Datei noch schliessen! */
  {
   INT32    len, rc;
   PFZGELEM pTab = p_FzgElemente;
   FILE *   file1;
   BYTE     b;

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->FzgNr ))
              return 1;
           if ( WriteInt32( file2, pTab->Len ))
              return 1;
           if ( WriteInt32( file2, pTab->Zeit ))
              return 1;
          }

        printf( pTab->PfName );
        printf( "\r\n" );

        file1 = OpenRead( pTab->PfName );
        if ( file1 == NULL )
           return 2;

        len = pTab->Len;
        while ( len-- )
          {
           rc = ReadByte( file1, &b );
           if ( rc )
              return 2;
           if ( ! Simulieren )
             {
              rc = WriteByte( file2, b );
              if ( rc )
                {
                 CloseRead( file1 );
                 return 1;
                }
             }
          }
        CloseRead( file1 );

        pTab = pTab->pNext;
       }

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL StoreNameGzg( PCHAR pPfad, PCHAR pName,
                                    INT32 Len, INT32 datum, INT32 zeit,
                                    BOOL Nutzerdef, PCHAR Archivname )
/* speichert Name der Einzeldatei und notwendige Eigenschaften
   (Landschaft)
   IN: pPfad = Pfad
       pName = Dateiname
       Len   = Laenge in BYTE
       datum = Bit0..4 Tag, Bit5..8 Monat, Bit9..15 Jahr seit 1980
       zeit  = Bit0..4 sec/2, Bit5..10 min, Bit11..15 h
       Nutzerdef = TRUE: nutzerdef. Datei
       Archivname = Dateinname des Archivs ohne Extension
                    (nur wichtig bei Nutzerdef == TRUE)
  OUT: 0 OK
       1 Problem beim Oeffnen/Datenlesen
       2 Speicher reicht nicht
       3 Datenfehler */
  {
   INT32    GSym, Bauform, sig, Phase, Alt, fw, scz, i, altzei;
   PGZGELEM pEl, pV, pN;
   PCHAR    pPos;
   BOOL     fahrweg;
   CHAR     VglName[ MAXPATH ];

     if ( Len <= 0 )
        return 1;

     fahrweg = FALSE;
     fw = 0;

     if ( Archivname != NULL )
        /* for case-insensitive compare */ 
        strlwr( Archivname );

     if ( ! Nutzerdef )
       {
        for ( i = 0; i < ANZAHL_FAHRWEGTYPEN; i++ )
          {
           if ( pName[0] == Fahrwegzeichen[i] )
             {
              fw = i + 1;
              fahrweg = TRUE;
              break;
             }
          }
       }
     else /* Nutzerdef, dann Dateiname filtern */
       {
        /* Diese Datei nicht aufnehmen, wenn der Name nicht enthalten ist
           (kein Fehler) */
        /* for case-insensitive compare */
        strcpy( VglName, pName );
        strlwr( VglName );
        SubExt( VglName );
        pPos = strstr( VglName+1, Archivname );
        if ( pPos == NULL )
           /* Archivname ist gar nicht drin */
           return 0;

        if ( strlen( pPos ) > strlen( Archivname ) )
           /* Archivname ist nur ein Teilstring */
           return 0;
        *pPos = '\0';
       }

     Phase = MIN_ANIPHASE;
     Alt   = KEINE_ALTERNATIVE;
     Bauform = sig = 0;

     if ( fahrweg )
       {
        sig = '\0'; /* %c writes LOBYTE only? */
        scz = sscanf( pName+1, "%u-%u%c", &GSym, &Bauform, &sig );
        if ( scz == 3 )
          {
           if (( sig == 'a' ) || ( sig == 'b' ))
              sig = sig - 'a' + 1; /* 1..2 */
           else
              sig = 0;
          }
        else
           sig = 0;

        if (( GSym < MIN_FAHRWEG ) || ( GSym > MAX_FAHRWEG ))
           return 3;
       }
     else
       {
        fw = sig = 0;

        altzei = '\0'; /* %c writes LOBYTE only? */
        if ( Nutzerdef )
           scz = sscanf( pName+1, "%u-%u%c", &GSym, &Phase, &altzei );
        else
           scz = sscanf( pName, "%u-%u%c", &GSym, &Phase, &altzei );
        if ( scz == 3 )
          {
           if (( altzei >= 'a' ) && ( altzei <= 'd' ))
              Alt = altzei - 'a' + 1;
           else
              Alt = KEINE_ALTERNATIVE;
          }
        else
          {
           Alt = KEINE_ALTERNATIVE;

           if ( scz == 1 )
             {
              Phase = MIN_ANIPHASE;
              /* dann nochmal ohne Phase versuchen */
              altzei = '\0'; /* %c writes LOBYTE only? */
              if ( Nutzerdef )
                 scz = sscanf( pName+1, "%u%c-", &GSym, &altzei );
              else
                 scz = sscanf( pName, "%u%c", &GSym, &altzei );
              if ( scz == 2 )
                {
                 if (( altzei >= 'a' ) && ( altzei <= 'd' ))
                    Alt = altzei - 'a' + 1;
                 else
                    Alt = KEINE_ALTERNATIVE;
                }
             }
          }

        if ( Nutzerdef )
          {
           if (( GSym < 0 ) || ( GSym >= USERGSYMZAHL_PER_DATEI ))
              return 3;
          }
        else
          {
           if (( GSym < MIN_LSYMBOL ) || ( GSym > MAX_LSYMBOL ))
              return 3;
          }
       }

  /* {
      CHAR8 xxx[120];
      sprintf( xxx, "sym=%i bf=%i fw=%i sig=%i azei=%c alt=%i ph=%i",
                    GSym, Bauform, fw, sig, altzei, Alt, Phase );
      Trace( xxx );
     } */

     if (( Phase < MIN_ANIPHASE ) || ( Phase > MAX_ANIPHASE ))
        return 3;

     if (( Alt < MIN_ALTERNATIVE ) || ( Alt > MAX_ALTERNATIVE ))
        if ( Alt != KEINE_ALTERNATIVE )
           return 3;

     pEl = AllocateMemory( sizeof( GZGELEM ) );
     if ( pEl == NULL )
        return 2;

     pEl->pNext   = NULL;
     pEl->SymNr   = GSym;
     pEl->Bauform = Bauform;
     pEl->FwSig   = (fw << 16) | sig;
     pEl->Phase   = Phase;
     pEl->Alt     = Alt;
     pEl->Len     = Len;
     pEl->Zeit  = (datum << 16) | zeit;
     strcpy( pEl->PfName, pPfad );
     strcat( pEl->PfName, "\\" );
     strcat( pEl->PfName, pName );

     /* InQueue/Einketten, sortiert */
     pV = NULL;
     pN = p_GzgElemente;

     while ( pN != NULL )
       {
        if ( pN->SymNr > GSym )
           break;

        if ( pN->SymNr == GSym )
          {
           if (( pN->Bauform > Bauform ) || ( pN->Phase > Phase ))
              break;

           if (( pN->Bauform == Bauform ) && ( pN->Phase == Phase ))
              if ( pN->Alt > Alt )
                 break;
          }
        pV = pN;
        pN = pN->pNext;
       }

     if ( pV == NULL )
       {
        pEl->pNext = p_GzgElemente;
        p_GzgElemente = pEl;
       }
     else
       {
        pEl->pNext = pV->pNext;
        pV->pNext = pEl;
       }

     Anzahl_Stored++;

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteSeekTableGzg( FILE * file2, BOOL Simulieren )
 /* Schreibt Tabelle der FzgNummern und seek()-Positionen
    (Fahrzeuge)
    IN:  file2 = Handle Datei
         Simulieren = TRUE: nichts wirklich schreiben
   OUT:  0 = OK
         1 = Schreibfehler (Datei geschlossen) */
  {
   INT32    len;
   PGZGELEM pTab = p_GzgElemente;

     /* Laenge der Seek-Tabelle selbst inkl. Endekz. */
     len = Anzahl_Stored * (5+1) * sizeof(INT32) + sizeof(INT32);

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->SymNr ))
              return 1;
           if ( WriteInt32( file2, pTab->Bauform ))
              return 1;
           if ( WriteInt32( file2, pTab->FwSig ))
              return 1;
           if ( WriteInt32( file2, pTab->Phase ))
              return 1;
           if ( WriteInt32( file2, pTab->Alt ))
              return 1;
          }

        /* abzgl. des eigenen Seek-Eintrags, der wurde ja dann schon gelesen */
        len -= 6*(sizeof(INT32));

        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, len ) )
              return 1;
          }

        /* Gesamtlaenge, + vor der Datei nochmal die Daten, eigene Laenge + Zeit
           (ist alles etwas redundant aber dadurch einfacher zu handhaben) */
        len += pTab->Len + (5+2) * sizeof( INT32 );

        pTab = pTab->pNext;
       }

     /* Endekennzeichen */
     if ( ! Simulieren )
       {
        if ( WriteInt32( file2, -1 ) )
           return 1;
       }    

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteDatenGzg( FILE * file2, BOOL Simulieren )
 /* Schreibt eigentlichen Datenteil, d.h. kopiert Dateien
    (Fahrzeuge)
    IN:  file2 = Handle Datei
         Simulieren = TRUE: nichts wirklich schreiben
   OUT:  0 = OK
         1 = Schreibfehler
         2 = Lesefehler
    Datei noch schliessen! */
  {
   INT32    len, rc;
   PGZGELEM pTab = p_GzgElemente;
   FILE *   file1;
   BYTE     b;

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->SymNr ))
              return 1;
           if ( WriteInt32( file2, pTab->Bauform ))
              return 1;
           if ( WriteInt32( file2, pTab->FwSig ))
              return 1;
           if ( WriteInt32( file2, pTab->Phase ))
              return 1;
           if ( WriteInt32( file2, pTab->Alt ))
              return 1;
           if ( WriteInt32( file2, pTab->Len ))
              return 1;
           if ( WriteInt32( file2, pTab->Zeit ))
              return 1;
          }

        printf( pTab->PfName );
        printf( "\r\n" );

        file1 = OpenRead( pTab->PfName );
        if ( file1 == NULL )
           return 2;

        len = pTab->Len;
        while ( len-- )
          {
           rc = ReadByte( file1, &b );
           if ( rc )
              return 2;
           if ( ! Simulieren )
             {
              rc = WriteByte( file2, b );
              if ( rc )
                {
                 CloseRead( file1 );
                 return 1;
                }
             }   
          }
        CloseRead( file1 );

        pTab = pTab->pNext;
       }

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL Auspacken( PCHAR FileName, PCHAR Directory,
                                 BOOL Ueberschreiben, BOOL Simulieren,
                                 PINT32 pFound,
                                 PINT32 pCopied, PINT32 pSkipped )
 /* Packt aus (Fahrzeuge oder Landschaft je nach Archivdatei)
    IN:  Filename = Archivdatei, kann Pfad enthalten
         Directory = Zielverzeichnis
         Ueberschreiben = TRUE: Zieldateien ueberschreiben wenn vorhanden
         Simulieren = TRUE: nichts wirklich schreiben
    OUT: 0 = OK, dann *pFound = Anzahl gefundene
                      *pCopied = Anzahl kopierte
                      *pSkipped = Anzahl uebersprungene
         1 = Archivdatei nicht gefunden
         2 = Lesefehler
         4 = Schreibfehler (Datentraeger voll?)
         5 = falscher Parameter */
  {
   FILE * file1, * file2;
   INT32  rc, fzgnr, pos, cnt, len, Zeit, i, anzahl, anzahlkop, anzahlueb;
   INT32  symnr, bauform, fwsig, phase, alt, Zoomfaktor;
   BYTE   b;
   CHAR8  Pfad[ MAXPATH ], PfadKpl[ MAXPATH ], Name[ 40 ];
   CHAR8  idbuf[ max( sizeof( IDENT_CODE_FZX ), sizeof( IDENT_CODE_GZX ) ) ];
   BOOL   Fahrzeuge, Nutzerdef, schreiben;
   CHAR8  Nutzerdateiname[ MAXPATH ];

     if ( strlen( FileName ) + 60 >= sizeof( Pfad ) )
        return 5;

     rc = fnsplit( FileName, NULL, NULL, Nutzerdateiname, NULL );
     if (( rc & WILDCARDS ) || !( rc & FILENAME ))
        return 5;

     anzahl = anzahlkop = anzahlueb = 0;
     if ( pFound != NULL )
        *pFound = 0;
     if ( pCopied != NULL )
        *pCopied = 0;
     if ( pSkipped != NULL )
        *pSkipped = 0;

     strcpy( Pfad, Directory );

     file1 = OpenRead( FileName );
     if ( file1 == NULL )
        return 1;

     /* Vorspann ueberlesen */
     do
       {
        rc = ReadByte( file1, &b );
        if ( rc )
           return 2;
       }
     while ( b != CTRL_Z );

     /* Ident. pruefen */
#if ( sizeof( IDENT_CODE_FZX ) != sizeof( IDENT_CODE_GZX ) )
#error So nicht!
#endif
     for ( i = 0; i < sizeof( idbuf ); i++ )
       {
        rc = ReadByte( file1, &b );
        if ( rc )
           return 2;
        idbuf[i] = b;
       }

     Fahrzeuge = Nutzerdef = FALSE;

     if ( memcmp( idbuf, IDENT_CODE_FZX, sizeof( idbuf ) ) == 0 )
        Fahrzeuge = TRUE;
     else
        if ( memcmp( idbuf, IDENT_CODE_UZX, sizeof( idbuf ) ) == 0 )
           Nutzerdef = TRUE;
        else
           if ( memcmp( idbuf, IDENT_CODE_GZX, sizeof( idbuf ) ) != 0 )
              return 2;

     rc = ReadByte( file1, &b );
     if ( rc )
        return 2;
     Zoomfaktor = b - '0';

     /* Seektabelle lesen
        (nur ueberlesen, es wird ja nichts gesucht) */
     while ( TRUE )
       {
        if ( Fahrzeuge )
          {
           rc = ReadInt32( file1, &fzgnr );
           if ( rc )
              return 2;
           if ( fzgnr == 0 )
              break;
          }
        else
          {
           rc = ReadInt32( file1, &symnr );
           if ( rc )
              return 2;
           if ( symnr == -1 )
              break;
           rc = ReadInt32( file1, &bauform );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &fwsig );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &phase );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &alt );
           if ( rc )
              return 2;
          }

        rc = ReadInt32( file1, &pos );
        if ( rc )
           return 2;

        anzahl++;
       }

     if ( ! Nutzerdef )
        Nutzerdateiname[0] = '\0';

     /* einzelne Dateien kopieren / copy single files */
     for ( cnt = 0; cnt < anzahl; cnt++ )
       {
        if ( Fahrzeuge )
          {
           rc = ReadInt32( file1, &fzgnr );
           if ( rc )
              return 2;
          }
        else
          {
           rc = ReadInt32( file1, &symnr );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &bauform );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &fwsig );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &phase );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &alt );
           if ( rc )
              return 2;
          }

        rc = ReadInt32( file1, &len );
        if ( rc )
           return 2;

        rc = ReadInt32( file1, &Zeit );
        if ( rc )
           return 2;

        strcpy( PfadKpl, Pfad );

        if ( Fahrzeuge )
          {
           sprintf( Name, "\\%05u-%02u.",
                    fzgnr >> 4, fzgnr & (FZGF_MAXFZGZAHL-1) );
           if ( Zoomfaktor == 2 )
              strcat( Name, FEXT_ZOOM2FZG );
           else
              strcat( Name, FEXT_ZOOM4FZG );
          }
        else
          {
           GetGzgDateiName( symnr, bauform, fwsig, alt, phase,
                            Name, Zoomfaktor, Nutzerdateiname );

        /* {
            CHAR8 xxx[120];
            sprintf( xxx, "sym=%i bf=%i fwsig=%xh alt=%i ph=%i n=%s",
                          symnr, bauform, fwsig, alt, phase, Name );
            Trace( xxx );
           } */
          }

        strcat( PfadKpl, "\\" );
        strcat( PfadKpl, Name );

        printf( PfadKpl );
        printf( " %s%i\r\n", GetText( TXTNR_012 ), len );

        if ( ! Ueberschreiben )
          {
           if ( DateiGibtEs( PfadKpl ) )
             {
              /* File already exists */
              printf( GetText( TXTNR_013 ) );
              schreiben = FALSE;
             }
           else
              schreiben = TRUE;
          }
        else
           schreiben = TRUE;

        /* bei ! Simulieren: dennoch schreiben = TRUE wegen Zaehlung/Test */

        if (( schreiben ) && ( ! Simulieren ))
          {
           file2 = OpenWrite( PfadKpl );
           if ( file2 == NULL )
              return 4;
          }

        while ( len-- )
          {
           rc = ReadByte( file1, &b );
           if ( rc )
             {
              if (( schreiben ) && ( ! Simulieren ))
                 CloseWrite( file2 );
              return 2;
             }
           if (( schreiben ) && ( ! Simulieren ))
             {
              rc = WriteByte( file2, b );
              if ( rc )
                {
                 CloseRead( file1 );
                 CloseWrite( file2 );
                 return 4;
                }
             }   
          }

        if (( schreiben ) && ( ! Simulieren ))
          {
           rc = CloseWrite( file2 );
           if ( rc )
             {
              CloseRead( file1 );
              return 4;
             }

           /* Datum + Zeit setzen
              dazu nochmal zum schreiben oeffnen
              (das letzte CloseWrite hat ggf. noch Puffer geschrieben, nach
               setftime darf aber kein schreiben mehr erfolgen) */
           if ( SetDateiAttr( PfadKpl, Zeit ) )
             {
              CloseRead( file1 );
              return 4;
             }

           anzahlkop++;
          } 
        else /* ! schreiben */
           anzahlueb++;
       }

     CloseRead( file1 );

     if ( pFound != NULL )
        *pFound = anzahl;
     if ( pCopied != NULL )
        *pCopied = anzahlkop;
     if ( pSkipped != NULL )
        *pSkipped = anzahlueb;

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL InfoUserEnd(VOID)
     /* Print Info text about parameters */
    {
     Meldung( GetText( TXTNR_006 ), GetText( TXTNR_005 ) );
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL Fehler( PCCHAR8 pText )
  {
     /* Print Error message */
     Meldung( pText, GetText( TXTNR_004 ) );
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL Trace( PCCHAR8 pText )
  {
     /* Print Trace message (for debug purposes only) */
     Meldung( pText, "Trace" );  /* language ENG always */
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL Meldung( PCCHAR8 pText, PCCHAR8 pTitel )
    {
     /* could be redirected to an output file or stdout/stderr... */
     MessageBox( NULL, pText, pTitel,
                 MB_OK | MB_ICONINFORMATION | MB_APPLMODAL );
    }

/* --------------------------------------------------------------------- */

STATIC CHAR FASTCALL Caps( CHAR zei )
     /* OUT: Grossbuchstabe wenn moegl. */
     {
      if (zei < 'a')
	 return zei;

      if (zei < ('z'+1))
	 return LOBYTE( zei - 32 );

      return zei;
     }

/*--------------------------------------------------------------------*/

VOID FASTCALL CapsString( PCHAR Text )
  /* wandelt Text in Grossbuchstaben um
     (auch mehrzeilig, Ende \0) */
  {
   INT32 i = 0;

     while ( Text[i] != '\0' )
       {
	Text[i] = Caps( Text[i] );
	i++;
       }
    }

/*--------------------------------------------------------------------*/

STATIC BOOL FASTCALL DateiGibtEs( PCCHAR8 pName )
   /* Prueft auf Vorhandensein einer Datei
      Checks whether file exists
      IN:   pName = Name der Datei, ggf. mit Pfad
      OUT:  TRUE: Datei existiert
            FALSE: nicht gefunden
            */
  {
   FILE * file;
   CHAR8 Buf[ MAXPATH ];
   INT32 len;

     if ( pName == NULL )
        return FALSE;
     len = strlen( pName );
     if (( len >= sizeof( Buf ) ) || ( len == 0 ))
        return FALSE;
     MemCopy1( Buf, pName, len+1 );

     file = fopen( Buf, "rb" );
     if ( file == NULL )
        return FALSE;

     fclose( file );
     return TRUE;
    }

/*--------------------------------------------------------------------*/

STATIC FILE * FASTCALL OpenWrite( PCCHAR8 pName )
   /* Oeffnet Datei zum Schreiben
      IN:   pName = Name der Datei
      OUT:  NULL Fehler (Datei nicht eroeffnet)
            sonst Handle
            */
  {
   FILE * file;

     file = fopen( pName, "wb" );
     if ( file == NULL )
        return NULL;

     WritePos = 0;
     return file;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL WriteByte( FILE * file2, BYTE b )
   /* schreibt Byte auf Ausgabedatei
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( WritePos < FILEBUFSIZE )
       {
        WriteBuf[ WritePos++ ] = b;
        return 0;
       }

     if ( fwrite( WriteBuf, FILEBUFSIZE, 1, file2 ) != 1 )
       {
        fclose( file2 );
        return 1;
       }

     WritePos = 1;
     WriteBuf[ 0 ] = b;

     return 0;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL WriteWord16( FILE * file2, WORD16 w )
   /* schreibt Word auf Ausgabedatei
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( WriteByte( file2, LOBYTE( w ) ) )
        return 1;

     return WriteByte( file2, HIBYTE( w ) );
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL WriteInt32( FILE * file2, INT32 w )
   /* schreibt INT32 auf Ausgabedatei
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( WriteWord16( file2, LOWORD( w ) ) )
        return 1;

     return WriteWord16( file2, HIWORD( w ) );
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL CloseWrite( FILE * file2 )
   /* schliet Ausgabedatei, mit Flush des restlichen Puffers
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler
            */
  {
   BOOL err = FALSE;

     if ( WritePos > 0 )
       {
        if ( fwrite( WriteBuf, WritePos, 1, file2 ) != 1 )
           err = TRUE;
       }

     if ( fclose( file2 ) != 0 )
        err = TRUE;

     if ( err )
        return 1;
     else
        return 0;
    }

/*--------------------------------------------------------------------*/

STATIC FILE * FASTCALL OpenRead( PCCHAR8 pName )
   /* Oeffnet Datei zum Lesen (gepuffert)
      IN:   pName = Name der Datei
      OUT:  NULL Fehler (Datei nicht eroeffnet)
            sonst Handle
            */
  {
   FILE * file;

     file = fopen( pName, "rb" );
     if ( file == NULL )
        return NULL;

     ReadPos = FILEBUFSIZE;
     return file;
    }

/*--------------------------------------------------------------------*/

INT32 FASTCALL ReadByte( FILE * file, PBYTE pB )
   /* liest Byte von Eingabedatei
      IN:   file = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( ReadPos < FILEBUFSIZE )
       {
        *pB = ReadBuf[ ReadPos++ ];
        return 0;
       }

     if ( fread( ReadBuf, 1, FILEBUFSIZE, file ) == 0 )
        /* am Ende der Datei ggf  rc < FILEBUFSIZE */
       {
        fclose( file );
        return 1;
       }

     ReadPos = 1;
     *pB = ReadBuf[ 0 ];

     return 0;
    }

/*--------------------------------------------------------------------*/

INT32 FASTCALL ReadWord16( FILE * file, PWORD16 pW )
   /* liest Word von Eingabedatei
      IN:   file = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
   BYTE bl;
   WORD16 w;

     if ( ReadByte( file, &bl ) )
        return 1;
     if ( ReadByte( file, (PBYTE) &w ) )
        return 1;

     *pW = LOWORD( ( w << 8 ) | bl );

     return 0;
    }

/*--------------------------------------------------------------------*/

INT32 FASTCALL ReadInt32( FILE * file, PINT32 pI )
   /* liest INT32 von Eingabedatei
      IN:   file = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
   WORD16 wl, wh;

     if ( ReadWord16( file, &wl ) )
        return 1;
     if ( ReadWord16( file, &wh ) )
        return 1;

     *pI = ( wh << 16 ) | wl;

     return 0;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL CloseRead( FILE * file )
   /* schliesst Eingabedatei
      IN:   file = FileHandle
      OUT:  0 OK
            1 Fehler
            */
  {
     if ( fclose( file ) != 0 )
        return 1;

     return 0;
    }

/* --------------------------------------------------------------------- */

PVOID FASTCALL AllocateMemory( INT32 len )
   /* allokiert Speicher
     IN:  len = Laenge in BYTE
    OUT:  NULL Fehler
          sonst Zeiger auf Block */
    {
     return malloc( len );
    }

/* --------------------------------------------------------------------- */

VOID FASTCALL FreeMemory( PVOID pDat )
   /* gibt allokierten Speicher frei
     IN:  pDat = Block */
    {
     free( pDat );
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL ReadOverSpacePos( PCCHAR8 pBuf )
 /* ueberliest Leerzeichen in Puffer pBuf
    IN:  pBuf = beliebiger Pufferbereich
    OUT: Offset erstes Nicht-' '
         (kann 0, d.h.  gleich der Eingabeposition sein) */
  {
   INT32 Pos = 0;

     while ( pBuf[ Pos ] == ' ' )
        Pos++;
     return Pos;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL DelSpaces( PCHAR8 pBuf )
 /* eliminiert Leerzeichen am Anfang des Puffers pBuf
    IN:  pBuf = beliebiger Pufferbereich  (Ende \0)
    OUT: pBuf = dasselbe, fuehrende Leerzeichen weggefallen
           rc = Anzahl geloeschte Leerzeichen */
  {
   INT32 Pos = ReadOverSpacePos( pBuf );
   INT32 Len;

     if ( Pos > 0 )
       {
        Len = strlen( pBuf + Pos );
        MemMove1( pBuf, pBuf + Pos, Len + 1 ); /* inkl. \0 */
       }

     return Pos;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL SetDateiAttr( PCCHAR8 pPfad, INT32 Zeit )
  /* setzt Datum und Uhrzeit
     IN:  pPfad = pPfad+Datei
          Zeit = HIWORD = Datum, LOWORD = Uhrzeit
                 im Format von findfirst()/findnext()
     OUT: 0 OK
          1 Datei nicht gefunden
          2 Schreibfehler  */
  {
   INT32 hfile, rc, rcc, datum, uhrzeit;
   struct ftime ftime;

     /* Datum + Zeit setzen
        dazu nochmal zum schreiben oeffnen
        (das letzte CloseWrite hat ggf. noch Puffer geschrieben, nach
         setftime darf aber kein schreiben mehr erfolgen) */
     hfile = open( pPfad, O_WRONLY | O_BINARY );
     if ( hfile == -1 )
        return 1;

     /* Bit0..4 Tag, Bit5..8 Monat, Bit9..15 Jahr seit 1980 */
     datum = Zeit >> 16;

     /* Bit0..4 sec/2, Bit5..10 min, Bit11..15 h */
     uhrzeit = LOWORD( Zeit );

     ftime.ft_tsec = uhrzeit & 0x001F;
     ftime.ft_min  = ( uhrzeit >> 5 ) & 0x003F;
     ftime.ft_hour = ( uhrzeit >> 11 ) & 0x001F;

     ftime.ft_day   = datum & 0x001F;
     ftime.ft_month = ( datum >> 5 ) & 0x000F;
     ftime.ft_year  = ( datum >> 9 ) & 0x007F;

     rc = setftime( hfile, &ftime );
     rcc = close( hfile );

     if (( rc == -1 ) || ( rcc == -1 ))
        return 2;
     else
        return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL GetSprache( VOID )
   /* Sprache ermitteln / Get Language
      OUT: SPRACHE_XXX */
  {
   LANGID sprachID;
   INT32  spcode, sprachIDPrim;

     /* get user language (Windows) */
     sprachID = GetUserDefaultLangID();

     /* primary ID is enough here: there are many sub-ID... */
     sprachIDPrim = PRIMARYLANGID( sprachID );
     switch( sprachIDPrim )
       {
        case LANG_GERMAN:
           spcode = SPRACHE_DEU;
           break;

        case LANG_SPANISH:
           spcode = SPRACHE_ESP;
           break;

        case LANG_CZECH:
        case LANG_SLOVAK: /* also Czech because similar */
           spcode = SPRACHE_CZE;
           break;

        case LANG_POLISH:
           spcode = SPRACHE_POL;
           break;

        default: /* including LANG_ENGLISH */
           spcode = SPRACHE_ENG;
       }

     return spcode;
    }

/* --------------------------------------------------------------------- */

STATIC PCCHAR8 FASTCALL GetText( INT32 Textnummer )
   /* Get text (depending on Sprach_Code)
      IN:  Textnummer = Text number
           Sprach_Code = current language
      OUT: pointer to the text */
  {
   STATIC  CHAR8 errtext[100]; /* STATIC wg. Rueckgabe / for giving back! */
   INT32   spcode, pos;
   PCCHAR8 pTxt;

/* Polskie tumaczenie / Polnische Uebersetzung: . M. P. P. 12.2008
   Czesky tumaczenie / Polnische Uebersetzung: . M. P. P. 12.2008
    (and P.Macku 1/2009)
   Remark JB about special characters:
    Correct coding Win/ANSI or DOS depends on purpose:
     Texts for console output: DOS Coding (here marked as "stdout")
           for messagebox output: Win/ANSI
     However: may be system dependent */

   STATIC CHAR8 txt00de[] = "Text fehlt:";
   STATIC CHAR8 txt00en[] = "Text missing:";
   STATIC CHAR8 txt00pl[] = "Brakuje tekstu:";
   STATIC CHAR8 txt00es[] = "Texto falta:";
   STATIC CHAR8 txt00cz[] = "Chyb text:";

   /* stdout */
   STATIC CHAR8 txt01de[] = "Version";
   STATIC CHAR8 txt01en[] = "Version";
   STATIC CHAR8 txt01pl[] = "Wersja";
   STATIC CHAR8 txt01es[] = "versin";
   STATIC CHAR8 txt01cz[] = "Verze";

   STATIC CHAR8 txt02de[] = "Archivdatei \"%s\" nicht erffnet";
   STATIC CHAR8 txt02en[] = "Archive file \"%s\" not opened";
   STATIC CHAR8 txt02pl[] = "Plik archiwum %s nie zosta otwarty";
   STATIC CHAR8 txt02es[] = "No se puede abrir el fichero de archivo \"%s\"";
   STATIC CHAR8 txt02cz[] = "Archivn soubor \"%s\" nebyl oteven";

   STATIC CHAR8 txt03de[] = "Archivdatei \"%s\" existiert schon";
   STATIC CHAR8 txt03en[] = "Archive file \"%s\" already exists";
   STATIC CHAR8 txt03pl[] = "Plik archiwum %s ju istnieje";
   STATIC CHAR8 txt03es[] = "El fichero \"%s\" ya existe";
   STATIC CHAR8 txt03cz[] = "Archivn soubor \"%s\" ji existuje";

   STATIC CHAR8 txt04de[] = "Fehler";
   STATIC CHAR8 txt04en[] = "Error";
   STATIC CHAR8 txt04pl[] = "Bd";
   STATIC CHAR8 txt04es[] = "Error";
   STATIC CHAR8 txt04cz[] = "Chyba";

   STATIC CHAR8 txt05de[] = "Korrekter Aufruf:";
   STATIC CHAR8 txt05en[] = "Correct call:";
   STATIC CHAR8 txt05pl[] = "Prawidowe wywoanie:";
   STATIC CHAR8 txt05es[] = "Usando correcto:";
   STATIC CHAR8 txt05cz[] = "Sprvn syntaxe:";

   STATIC CHAR8 txt06de[] =
"FZGC2 Kommando Verzeichnis Dateiname [Optionen]\n\n\
Kommandos: V2|V4|S1|S2|S4|U1|U2|U4|X\n\
V2: \"Verzeichnis\\*.fz2\" => \"Dateiname.fz2\"\n\
V4: \"Verzeichnis\\*.fz4\" => \"Dateiname.fz4\"\n\
Sn: \"Verzeichnis\\*.gzn\" => \"Dateiname\"\n\
Un: \"Verzeichnis\\u??-???-Dateiname.gzn\" => \"Dateiname[.uzn]\"\n\
X: in \"Dateiname\" enthaltene Dateien => \"Verzeichnis\\*.*\"\n\n\
Optionen: [O][S]\n\
O: berschreiben existierender Dateien\n\
S: Simulationsmodus (nichts schreiben)\n\n\
Keine Leerzeichen in Namen\n\
Kommandos/Optionen Gro-/Kleinschreibung egal";
   STATIC CHAR8 txt06en[] =
"FZGC2 command directory filename [options]\n\n\
Commands: V2|V4|S1|S2|S4|U1|U2|U4|X\n\
V2: \"directory\\*.fz2\" => \"filename.fz2\"\n\
V4: \"directory\\*.fz4\" => \"filename.fz4\"\n\
Sn: \"directory\\*.gzn\" => \"filename\"\n\
Un: \"directory\\u??-???-filename.gzn\" => \"filename[.uzn]\"\n\
X: in \"filename\" stored files => \"directory\\*.*\"\n\n\
Options: [O][S]\n\
O: Overwrite existing files\n\
S: Simulation mode (write nothing)\n\n\
No spaces in names\n\
Commands/Options case insensitive";
   STATIC CHAR8 txt06pl[] =
"FZGC2 polecenie cieka nazwa_pliku [opcje]\n\n\
Polecenia: V2|V4|S1|S2|S4|U1|U2|U4|X\n\
V2: cieka\\*.fz2 => nazwa_pliku.fz2\n\
V4: cieka\\*.fz4 => nazwa_pliku.fz4\n\
Sn: cieka\\*.gzn => nazwa_pliku\n\
Un: cieka\\u??-???-nazwa_pliku.gzn => nazwa_pliku[.uzn]\n\
X: pliki zawarte w nazwa_pliku => cieka\\*.*\n\n\
Opcje: [O][S]\n\
O: Nadpisuj istniejce pliki\n\
S: Tryb symulacji (bez zapisu)\n\n\
Nazwy plikw/katalogw nie mog zawiera spacji\n\
Wielko liter nie jest uwzgldniana w poleceniach/opcjach";
   STATIC CHAR8 txt06es[] =
"FZGC2 comando directorio nombre_de_fichero [opciones]\n\n\
Comandos: V2|V4|S1|S2|S4|U1|U2|U4|X\n\
V2: \"directorio\\*.fz2\" => \"nombre.fz2\"\n\
V4: \"directorio\\*.fz4\" => \"nombre.fz4\"\n\
Sn: \"directorio\\*.gzn\" => \"nombre\"\n\
Un: \"directorio\\u??-???-nombre.gzn\" => \"nombre[.uzn]\"\n\
X: ficheros guardados en \"nombre\" => \"directorio\\*.*\"\n\n\
Opciones: [O][S]\n\
O: reemplazar ficheros existente\n\
S: modo de la simulacin (no escribir nada)\n\n\
No espacios en nombres\n\
Comandos/opciones insignificantes si con mayscula/minscula";
   STATIC CHAR8 txt06cz[] =
"FZGC2 pkaz cesta nzev_souboru [volby]\n\n\
Pkazy: V2|V4|S1|S2|S4|U1|U2|U4|X\n\
V2: \"cesta\\*.FZ2\" => \"nzev_souboru.FZ2\"\n\
V4: \"cesta\\*.FZ4\" => \"nzev_souboru.FZ4\"\n\
Sn: \"cesta\\*.GZn\" => \"nzev_souboru\"\n\
Un: \"cesta\\u??-???-nzev_souboru.GZn\" => \"nzev_souboru[.UZn]\"\n\
X: soubory obsaen v \"nzev_souboru\" => \"cesta\\*.*\"\n\n\
Volby: [O][S]\n\
O: Pepsat existujc soubory\n\
S: Md simulace (bez zpisu)\n\n\
Nzvy soubor a sloek nesm obsahovat mezery\n\
Pkazy a volby nerozliuj velikost psmen.";

   STATIC CHAR8 txt07de[] =
"Dateien gefunden: %i  ausgepackt: %i  bersprungen: %i";
   STATIC CHAR8 txt07en[] =
"Files found: %i  extracted: %i  skipped: %i";
   STATIC CHAR8 txt07pl[] =
"Znaleziono plikw: %i  wypakowano: %i  pominito: %i";
   STATIC CHAR8 txt07es[] =
"Ficheros encontrados: %i  generados: %i  cancelados: %i";
   STATIC CHAR8 txt07cz[] =
"Nalezeno soubor: %i  Rozbaleno: %i  Peskoeno: %i";

   /* multiple used! */
   STATIC CHAR8 txt08de[] = "Ende/fertig";
   STATIC CHAR8 txt08en[] = "Done/ready";
   STATIC CHAR8 txt08pl[] = "Gotowe";
   STATIC CHAR8 txt08es[] = "Terminado";
   STATIC CHAR8 txt08cz[] = "Hotovo";

   STATIC CHAR8 txt09de[] = "Auspacken: rc=";
   STATIC CHAR8 txt09en[] = "Extracting: rc=";
   STATIC CHAR8 txt09pl[] = "Wypakowywanie: kod zwrotny=";
   STATIC CHAR8 txt09es[] = "Descomprimir: rc=";
   STATIC CHAR8 txt09cz[] = "Rozbalovn: rc=";

   STATIC CHAR8 txt10de[] = "Eingepackt: %i Dateien";
   STATIC CHAR8 txt10en[] = "Packed: %i files";
   STATIC CHAR8 txt10pl[] = "Spakowano: %i plikw";
   STATIC CHAR8 txt10es[] = "Comprimiendos: %i ficheros";
   STATIC CHAR8 txt10cz[] = "Sbaleno: %i soubor";

   STATIC CHAR8 txt11de[] = "Einpacken: rc=";
   STATIC CHAR8 txt11en[] = "Packing: rc=";
   STATIC CHAR8 txt11pl[] = "Pakowanie: kod zwrotny=";
   STATIC CHAR8 txt11es[] = "Comprimir: rc=";
   STATIC CHAR8 txt11cz[] = "Balen: rc=";

   /* stdout */
   STATIC CHAR8 txt12de[] = "Laenge=";
   STATIC CHAR8 txt12en[] = "Length=";
   STATIC CHAR8 txt12pl[] = "Dugo=";
   STATIC CHAR8 txt12es[] = "La longitud=";
   STATIC CHAR8 txt12cz[] = "Dlka="; /* Delka */

   /* stdout */
   STATIC CHAR8 txt13de[] = "Datei existiert schon\n";
   STATIC CHAR8 txt13en[] = "File already exists\n";
   STATIC CHAR8 txt13pl[] = "Plik ju istnieje\n";
   STATIC CHAR8 txt13es[] = "El fichero ya existe\n";
   STATIC CHAR8 txt13cz[] = "Soubor ji existuje\n"; /* S. jiz e. */

   STATIC CHAR8 txt14de[] = "Keine Datei gefunden";
   STATIC CHAR8 txt14en[] = "No file found";
   STATIC CHAR8 txt14pl[] = "Plik nie zosta znaleziony";
   STATIC CHAR8 txt14es[] = "Ninguno fichero no encuentro";
   STATIC CHAR8 txt14cz[] = "Soubor nenalezen";

   STATIC CHAR8 txt15de[] = "Speicher reicht nicht";
   STATIC CHAR8 txt15en[] = "Out of memory";
   STATIC CHAR8 txt15pl[] = "Zabrako pamici";
   STATIC CHAR8 txt15es[] = "No hay memoria";
   STATIC CHAR8 txt15cz[] = "Nedostatek pamti";

   STATIC CHAR8 txt16de[] = "Keine Archivdatei erzeugt";
   STATIC CHAR8 txt16en[] = "No archive file created";
   STATIC CHAR8 txt16pl[] = "Plik archiwum nie zosta utworzony";
   STATIC CHAR8 txt16es[] = "Ninguno fichero de archivo no generado";
   STATIC CHAR8 txt16cz[] = "Archivn soubor nebyl vytvoen";

   STATIC CHAR8 txt17de[] = "Schreibfehler";
   STATIC CHAR8 txt17en[] = "Error while writing";
   STATIC CHAR8 txt17pl[] = "Bd podczas zapisu";
   STATIC CHAR8 txt17es[] = "Error al escribir";
   STATIC CHAR8 txt17cz[] = "Chyba bhem zpisu";

   STATIC CHAR8 txt18de[] = "Lesefehler";
   STATIC CHAR8 txt18en[] = "Error while reading";
   STATIC CHAR8 txt18pl[] = "Bd odczytu";
   STATIC CHAR8 txt18es[] = "Error de lectura";
   STATIC CHAR8 txt18cz[] = "Chyba ten souboru";

   STATIC CHAR8 txt19de[] = "Datenfehler";
   STATIC CHAR8 txt19en[] = "Data error";
   STATIC CHAR8 txt19pl[] = "Bd danych";
   STATIC CHAR8 txt19es[] = "Error de datos";
   STATIC CHAR8 txt19cz[] = "Chyba ten dat";

   CONST TXTELEM pTxte[] = {
    /* Order like Text numbers (but not needed) and SPRACHE_xxx (needed)
       End marker = TXTNR_UNDEF */
     { TXTNR_000, { txt00en, txt00de, txt00pl, txt00es, txt00cz }},
     { TXTNR_001, { txt01en, txt01de, txt01pl, txt01es, txt01cz }},
     { TXTNR_002, { txt02en, txt02de, txt02pl, txt02es, txt02cz }},
     { TXTNR_003, { txt03en, txt03de, txt03pl, txt03es, txt03cz }},
     { TXTNR_004, { txt04en, txt04de, txt04pl, txt04es, txt04cz }},
     { TXTNR_005, { txt05en, txt05de, txt05pl, txt05es, txt05cz }},
     { TXTNR_006, { txt06en, txt06de, txt06pl, txt06es, txt06cz }},
     { TXTNR_007, { txt07en, txt07de, txt07pl, txt07es, txt07cz }},
     { TXTNR_008, { txt08en, txt08de, txt08pl, txt08es, txt08cz }},
     { TXTNR_009, { txt09en, txt09de, txt09pl, txt09es, txt09cz }},
     { TXTNR_010, { txt10en, txt10de, txt10pl, txt10es, txt10cz }},
     { TXTNR_011, { txt11en, txt11de, txt11pl, txt11es, txt11cz }},
     { TXTNR_012, { txt12en, txt12de, txt12pl, txt12es, txt12cz }},
     { TXTNR_013, { txt13en, txt13de, txt13pl, txt13es, txt13cz }},
     { TXTNR_014, { txt14en, txt14de, txt14pl, txt14es, txt14cz }},
     { TXTNR_015, { txt15en, txt15de, txt15pl, txt15es, txt15cz }},
     { TXTNR_016, { txt16en, txt16de, txt16pl, txt16es, txt16cz }},
     { TXTNR_017, { txt17en, txt17de, txt17pl, txt17es, txt17cz }},
     { TXTNR_F_LESEN, { txt18en, txt18de, txt18pl, txt18es, txt18cz }},
     { TXTNR_F_DATEN, { txt19en, txt19de, txt19pl, txt19es, txt19cz }},
#if ( SPRACHE_ANZAHL > 5 )
#error Add some NULLs!
#endif      
     { TXTNR_UNDEF, { NULL, NULL, NULL, NULL, NULL }}
     };

    if (( Sprach_Code < SPRACHE_MIN ) || ( Sprach_Code > SPRACHE_MAX ))
       spcode = SPRACHE_ENG;
    else
       spcode = Sprach_Code;

    pos = 0;
    while ( pTxte[pos].TxtNr != TXTNR_UNDEF )
      {
       if ( pTxte[pos].TxtNr == Textnummer )
          break;
       pos++;
      }

    pTxt = pTxte[pos].pTxt[spcode];
    if ( pTxt == NULL )
      {
       sprintf( errtext, "%s %i",
                pTxte[0].pTxt[spcode], Textnummer );
       return errtext;
      }

    return pTxt;
   }

/*--------------------------------------------------------------------*/

STATIC VOID FASTCALL MemMove1( PVOID pDest, PCVOID pSrc, INT32 len )
/* same as memmove()
   copies byte[len] from pSrc to pDest
   with correct handling of overlapping areas */
    {
     if (( pDest == NULL ) || ( pSrc == NULL ) || ( len <= 0 ))
        return;

#ifdef TEST310
     if ( len >= 32767 )
       {
        CHAR8 bbb[80];
        printf( bbb, "MemMove1 len = %u ", len );
        Trace( bbb );
        return;
       }
     if ( pDest == pSrc )
       {
        CHAR8 bbb[80];
        printf( bbb, "MemMove1 Dest==Src " );
        Trace( bbb );
        return;
       }
#endif

     memmove( pDest, pSrc, len );
    }

/*------------------------------------------------------- */

STATIC INT32 FASTCALL GetGzgDateiName( INT32 GSym, INT32 Bauform, INT32 FwSig,
                                       INT32 Alt, INT32 Phase,
                                       PCHAR8 pFName, INT32 Zf,
                                       PCCHAR8 pNutzerName )
/* liefert Dateiname fuer einzelne Grafikdatei zu einem Landschaftssymbol
  IN: GSym = Landschaftssymbolnr/nutzerdef.Symbolnr./Fahrwegsymbol
      Bauform = bei Fahrweg: Bauform
      FwSig   = bei Fahrweg: HIWORD: Fahrweginfo
                             LOWORD: bei Signalen Phase 1/2
      Alt = bei Landschaft: KEINE_ALTERNATIVE = es gibt nur diese Variante
                            MIN_ALTERNATIVE..MAX_ = diese ist gemeint
      Phase   = Phase wenn Animation (0 wenn keine)
      pFName = Rueckgabepuffer [MAXPATH]
      Zf = Zoomfaktor (1/2/4)
      pNutzerName = Name der nutzerdef. Datei oder leer bei Standard
 OUT: 0 OK, Standardsymbol (entspr. Standard-Pfad oder Sammeldatei benutzen)
      1 OK, nutzerdef.Symbol (Datenpfad benutzen)
      2 falsche Eingangsdaten (Symbolnr./Zf)
	    */
  {
   INT32   AddZeichen, FwZei, SigZustZei;
   CHAR8   PhaseBuf[20], BauformBuf[20];
   PCHAR8  pExt;
   BOOL    nutzerdef, fahrweg;

     if ( pFName == NULL )
        return 2;
     pFName[0] = '\0'; /* erstmal immer */

     if ( pNutzerName[0] != '\0' )
        /* nutzerdefiniert */
        nutzerdef = TRUE;
     else
        nutzerdef = FALSE;

     if (( GSym <= MAX_FAHRWEG ) && ( ! nutzerdef ))
        fahrweg = TRUE;
     else
        fahrweg = FALSE;


     if ( fahrweg )
       {
        Alt   = KEINE_ALTERNATIVE;
        Phase = MIN_ANIPHASE;
       }
     else
       {
        Bauform = 0;
       }

     if ( Zf == 1 )
        pExt = FEXT_ZOOM1GZG;
     else
        if ( Zf == 2 )
           pExt = FEXT_ZOOM2GZG;
        else
           if ( Zf == 4 )
              pExt = FEXT_ZOOM4GZG;
           else
              return 2;

     if (( Alt >= MIN_ALTERNATIVE ) && ( Alt <= MAX_ALTERNATIVE ))
        AddZeichen = 'a'-1 + Alt;  /* 1..4 => a..d */
     else
        AddZeichen = '\0';

     if (( Phase > MIN_ANIPHASE ) && ( Phase <= MAX_ANIPHASE ))
       {
        sprintf( PhaseBuf, "-%02i", Phase );
       }
     else
        PhaseBuf[0] = '\0';

     /* Bauform syntaktisch wie Phase, aber das eine nur Fahrweg,
        das andere nur Landschaft */
     if ( fahrweg )
       {
        if ( LOWORD( FwSig ) == 0 )
           SigZustZei = '\0';
        else /* Signal */
          {
           SigZustZei = 'a' + LOWORD( FwSig ) - 1;
          }

        sprintf( BauformBuf, "-%03i%c", Bauform, SigZustZei );

        FwZei = Fahrwegzeichen[ HIWORD( FwSig )-1 ];

        /* Dateiname aus Symbolnummer bilden
           (unabhaengig vom Fahrweg) */
        sprintf( pFName, "%c%03u",
                         FwZei, GSym );

       }
     else /* kein Fahrweg */
       {
        BauformBuf[0] = '\0';

        if ( nutzerdef )
           /* nutzerdefiniert */
          {
           /* Symbole 0..71 */
           sprintf( pFName, "%c%02u",
                            ZEICHEN_NUTZERDEF, GSym );
          }
        else /* Standard */
          {
           /* Dateiname aus Symbolnummer bilden */
           sprintf( pFName, "%06u", GSym );
          }
       }

     if ( PhaseBuf[0] != '\0' )
        strcat( pFName, PhaseBuf );

     if ( BauformBuf[0] != '\0' )
        strcat( pFName, BauformBuf );

     if ( AddZeichen != '\0' )
        strcadd( pFName, LOBYTE( AddZeichen ));

     if ( nutzerdef )
       {
        strcadd( pFName, '-' );
        strcat( pFName, pNutzerName );
       }

     strcadd( pFName, '.' );
     strcat( pFName, pExt );

     if ( nutzerdef )
        return 1;
     else
        return 0;
    }

/*------------------------------------------------------- */

STATIC VOID FASTCALL strcadd( PCHAR pStr, CHAR Zei )
    /* haengt ein Zeichen an den String an */
  {
   CHAR str[2];

     str[0] = Zei;
     str[1] = '\0';
     strcat( pStr, str );
    }

/*------------------------------------------------------- */

STATIC VOID FASTCALL SubExt( PCHAR pName )
 /* loescht Extension, wenn vorhanden
    Unter Beruecksichtigung, dass Vz.name einen '.' enthalten kann */
  {
   INT32  p = strlen( pName ) - 1;

     while (( p > 0 ) && ( pName[p] != '\\' ))
       {
        if ( pName[p] == L'.' )
          {
           pName[p] = '\0';
           return;
          }
        else
           p--;
       }
    }

