/* JB T14W Project (BAHN) - FZGC2 Graphics Compiler

 Jan Bochmann 2/2007

 Stand/Last changes:  2011-02-28 (for BAHN 3.86Beta2)

DE: Compiler zum Erstellen von Archivdateien aus einzelnen Dateien
    sowie umgekehrt zum Erstellen selbiger aus dem jeweiligen Archiv

EN: Compiler for creating archive files from single files
    and back for creating them from the respective archive files

    ".fz2" => "bahn.fz2", "name.fz2",
    ".fz4" => "bahn.fz4", "name.fz4",
    ".gz1" => "bahn.fw1", "bahn.fx1", "name.uz1"
    ".gz2" => "bahn.fw2", "bahn.fx2", "name.uz2"
    ".gz4" => "bahn.fw4", "bahn.fx4", "name.uz4"

DE: Aufrufparameter: Werden bei Start ohne Parameter angezeigt
    Leerzeichen ist Trennzeichen, d.h. Namen, die Leerzeichen enthalten,
    werden nicht unterstuetzt.

EN: Arguments for call: Are displayed when starting without parameters
    Empty space is separator, i.e. names containing spaces are not supported.
   */

/* ------------------------------------------------------------------------ */

/* DE: Dies ist eigtl. ein Konsolenprogramm und damit portierbar,
       aber folgende Windows-Funktionen werden verwendet:
   EN: Program is made for console/command line and in result portable,
       however the following MS-Windows functions are used:

       MessageBox()
       GetUserDefaultLangID()
       GlobalAlloc(), GlobalFree()  (optionally)
            */

#include <windows.h>

#pragma hdrstop

/* ------------------------------------------------------------------------
DE: Format der Ausgabedateien:
    (Aufbau der Einzeldateien siehe extra Dokumentation)

  Zoom2 Fahrzeuge:

  1. BYTE[] Text Vorspann, ANSI, Ende 'CTRL_Z', '\0'
  2. BYTE IDENT_CODE[4]={'F','Z','X','2'}
  3. Versionskennung
      INT32 = ZOOMXARCHIVEFZGFORMAT_386_0
  4. Tabelle fr schnelleres Suchen mit lseek()
     je Eintrag:
      INT32  Fahrzeugsatznummer
             ( -1 == Ende, danach keine Fzg.nummer und kein Offset )
      INT32  Fahrzeugnummer im Satz
      INT32  Dateioffset fr lseek() ab Position hinter dem INT32
             zum Erreichen der Daten des Fahrzeugs
  5. Fahrzeuge:
     je Fahrzeug:
      INT32    Fzg.satznummer (dieselbe wie in Tabelle 4.)
      INT32    Fzg.nummer (dieselbe wie in Tabelle 4.)
      INT32    Laenge der Einzeldatei in BYTE
      INT32    Datum/Zeit der Einzeldatei
      BYTE[]   komplette Einzeldatei .fz2 (binaer kopiert)
  6. keine Endekennung

    Die Fahrzeuge und Tabelleneintraege sind sortiert von unten nach oben.

  Zoom4 Fahrzeuge:

   Identisch zu Zoom2, auer:

   2. BYTE IDENT_CODE[4]={'F','Z','X','4'}

  Zoom1/2/4 Landschaft, nutzerdefiniert:

  1. Vorspann bis CTRL Z
  2. BYTE IDENT_CODE[3] ={'U','Z','X'}
  3. BYTE ZoomFaktorID = '1' oder '2' oder '4'
  4. Vorspann fr schnelleres Suchen mit lseek()
      INT32  Elementnummer lokal in der Datei 0..89
                 (-1 == Endekennzeichen, danach keine weiteren Daten)
      INT32  Bauform = 0
      INT32  FwSig = 0
      INT32  Phase = Animationsphase
      INT32  Alt   = Alternative (0 oder 1..4 fuer "a".."d")
      INT32  Dateioffset fr lseek() ab Position hinter der Elementnummer
             zum Erreichen des Datenblocks
  5. Elemente (Datenbloecke):
      INT32  Elementnummer (die selbe wie im Vorspann)
      INT32  Bauform (die selbe wie im Vorspann)
      INT32  FwSig (die selbe wie im Vorspann)
      INT32  Phase (die selbe wie im Vorspann)
      INT32  Alt   (die selbe wie im Vorspann)
      INT32  Laenge der Einzeldatei in BYTE
      INT32  Datum/Zeit der Datei
      BYTE[] komplette .gz1/2/4-Datei (binaer kopiert)
  6. keine Endekennung

  Zoom1/2/4 Standard (Landschaft und Fahrwegdaten):

  1. Vorspann bis CTRL Z
  2. BYTE IDENT_CODE[3] ={'G','Z','X'}
  3. BYTE ZoomFaktorID = '1' oder '2' oder '4'
  4. Vorspann fr schnelleres Suchen mit lseek()
      INT32  Elementnummer (global in BAHN)
                (-1 == Endekennzeichen, danach keine weiteren Daten)
      INT32  Bauform (bei Landschaft immer = 0)
      INT32  FwSig: bei Landschaft = 0
                    bei Fahrweg: HIWORD = Fahrweg 1..4 (FAHRWEGTYP_xxx+1)
                                 LOWORD = bei Signalen Signalphase
                                          1/2/3/4 fuer "a"/"b"/"c"/"d"
                                          sonst = 0
      INT32  Phase = Animationsphase (bei Fahrweg immer = 0)
      INT32  Alt   = Alternative (0 oder 1..4 fuer "a".."d")
                     bei Fahrweg immer = 0
      INT32  Dateioffset fr lseek() ab Position hinter der Elementnummer
             zum Erreichen des Datenblocks
  5. Elemente (Datenbloecke):
      INT32  Elementnummer (die selbe wie im Vorspann)
      INT32  Bauform (die selbe wie im Vorspann)
      INT32  FwSig (die selbe wie im Vorspann)
      INT32  Phase (die selbe wie im Vorspann)
      INT32  Alt   (die selbe wie im Vorspann)
      INT32  Laenge der Einzeldatei in BYTE
      INT32  Datum/Zeit der Datei
      BYTE[] komplette .gz1/2/4-Datei (binaer kopiert)
  6. keine Endekennung

EN: Output file format:
    (for structure of single files see extra documentation)

  Zoom2 Vehicles:

  1. BYTE[] Text header, ANSI, terminated by 'CTRL_Z', '\0'
  2. BYTE IDENT_CODE[4]={'F','Z','X','2'}
  3. Version identifier
      INT32 = ZOOMXARCHIVEFZGFORMAT_386_0
  4. Table for faster search using lseek()
     for each entry:
      INT32  Vehicle set number
             ( -1 == terminator, after neither car number nor offset)
      INT32  Vehicle number in the set
      INT32  File offset for lseek() from position behind the INT32
             for reaching the vehicle data of this vehicle
  5. Vehicles:
     for each:
      INT32    Vehicle set number (same as in table 4.)
      INT32    Vehicle number (same as in table 4.)
      INT32    Length of single file in BYTE
      INT32    Date/time of single file
      BYTE[]   Complete single file .fz2 (binary copied)
  6. no terminator

    The vehicles and table entries are sorted from smallest to highest.

  Zoom4 Vehicles:

   Identical to Zoom2, except:

   2. BYTE IDENT_CODE[4]={'F','Z','X','4'}

  Zoom1/2/4 Scenery, user-defined:

  1. Header (text) until CTRL Z
  2. BYTE IDENT_CODE[3] ={'U','Z','X'}
  3. BYTE ZoomFaktorID = '1' or '2' or '4'
  4. Header for faster search using lseek()
      INT32  Element number locally in the file 0..89
                 (-1 == terminator, after it no data)
      INT32  Bauform = 0
      INT32  FwSig = 0
      INT32  Phase = Animation phase
      INT32  Alt   = Alternative/Variant (0 or 1..4 for "a".."d")
      INT32  file offset for lseek() from position behind the element number
             for reaching the data block
  5. Elements (data blocks):
      INT32  Element number (same as in header)
      INT32  Bauform (same as in header)
      INT32  FwSig (same as in header)
      INT32  Phase (same as in header)
      INT32  Alt   (same as in header)
      INT32  length of single file in BYTE
      INT32  date/time of single file
      BYTE[] complete .gz1/2/4 file (binary copied)
  6. no terminator

  Zoom1/2/4 Standard (scenery and driving way add-on):

  1. Header (text) until CTRL Z
  2. BYTE IDENT_CODE[3] ={'G','Z','X'}
  3. BYTE ZoomFaktorID = '1' or '2' or '4'
  4. Header for faster search using lseek()
      INT32  Element number (globally in BAHN)
                (-1 == terminator, after it no data)
      INT32  Bauform: for scenery always = 0
                      for driving way: graphics version
      INT32  FwSig: for scenery = 0
                    for driv.way: HIWORD = dr.way 1..4 (FAHRWEGTYP_xxx+1)
                                  LOWORD = for signals signal phase
                                           1/2/3/4 for "a"/"b"/"c"/"d"
                                           otherwise = 0
      INT32  Phase = Animation phase (for driv.way always = 0)
      INT32  Alt   = Alternative/Variant (0 or 1..4 for "a".."d")
                     for driving way = 0
      INT32  file offset for lseek() from position behind the element number
             for reaching the data block
  5. Elements (data blocks):
      INT32  Element number (same as in header)
      INT32  Bauform (same as in header)
      INT32  FwSig (same as in header)
      INT32  Phase (same as in header)
      INT32  Alt   (same as in header)
      INT32  length of single file in BYTE
      INT32  date/time of single file
      BYTE[] complete .gz1/2/4 file (binary copied)
  6. no terminator

 */

 /* See AllocateMemory() and FreeMemory()
    if def then use malloc() and free()
    else use Windows functions */
#undef USE_MEM_STANDARD_LIBRARY

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <dir.h>
#include <string.h>
#include <fcntl.h>
#ifdef USE_MEM_STANDARD_LIBRARY
#include <alloc.h>
#endif

#define FASTCALL        _fastcall
#define CONST           const
#define STATIC          static
/* typedef void VOID; */
typedef unsigned char   BYTE;
typedef   signed char   CHAR8;
typedef   signed short  INT16;
typedef   signed short  CHAR16;
typedef unsigned short  WORD16;
typedef   signed int    INT32;
typedef unsigned int    WORD32;
typedef int             BOOL;
typedef VOID *          PVOID;
typedef VOID CONST *    PCVOID;
typedef BYTE *          PBYTE;
typedef CHAR8 *         PCHAR8;
typedef CHAR8 CONST *   PCCHAR8;
typedef CHAR16 *        PCHAR16;
typedef CHAR16 CONST *  PCCHAR16;
typedef WORD16 *        PWORD16;
typedef INT16 *         PINT16;
typedef WORD32 *        PWORD32;
typedef INT32 *         PINT32;
typedef BOOL *          PBOOL;
#define TRUE  1
#define FALSE 0

/* #define LOWORD(l)	    ((WORD16)(l))
#define HIWORD(l)	    ((WORD16) ((l) >> 16))
#define LOBYTE(w)	    ((w) & 0xFF)
#define HIBYTE(w)	    ((w) >> 8) */

#define CTRL_Z  26

/* Testuntersttzung */
#undef TEST10
#undef TEST20
#undef TEST201
#undef TEST310

/* Sprachen / Languages (not all supported) */
#define SPRACHE_MIN         0
#define SPRACHE_ENG         0
#define SPRACHE_DEU         1
#define SPRACHE_POL         2
#define SPRACHE_ESP         3
#define SPRACHE_CZE         4
#define SPRACHE_MAX         4
#define SPRACHE_ANZAHL      (SPRACHE_MAX-SPRACHE_MIN+1)
/* #define SPRACHE_FRA         2 */
/* #define SPRACHE_NLD         3 */

/* Global variables */
CONST CHAR8 IDENT_CODE_FZX[3] = { 'F', 'Z', 'X' };
CONST CHAR8 IDENT_CODE_GZX[3] = { 'G', 'Z', 'X' };
CONST CHAR8 IDENT_CODE_UZX[3] = { 'U', 'Z', 'X' };

/* File extensions single files */
#define FEXT_ZOOM2FZG  "fz2"  /* b384,b385beta1..3: "fzz" */
#define FEXT_ZOOM4FZG  "fz4"
#define FEXT_ZOOM1GZG  "gz1"
#define FEXT_ZOOM2GZG  "gz2"
#define FEXT_ZOOM4GZG  "gz4"
/* File extensions archive files (+ '0'+zoomfaktor) */
#define FEXT_ARCZOOMFZG  ".fz"
#define FEXT_ARCZOOMGZG  ".gz"
#define FEXT_ARCZOOMUZG  ".uz"

/* File versions for single files
   see BAHN gfx documentation for details */

#define ZOOM2FZGFORMAT383  0x0383
#define ZOOM2FZGSUBFORMAT2   0x02
#define ZOOM2FZGSUBFORMAT3   0x03

#define ZOOMXFZGFORMAT385  0x0385
#define ZOOMXFZGSUBFORMAT0   0x00
#define ZOOMXFZGSUBFORMAT3   0x03

/* File versions for archive files
   see BAHN gfx documentation for details */
#define ZOOMXARCHIVEFZGFORMAT_386_0   (0x03860000)
#define ZOOMXARCHIVEFZGFORMAT_386_alt (ZOOMXARCHIVEFZGFORMAT_386_0 - 1)

/* Vehicle properties
   see BAHN gfx documentation for details */

#define ZGGEIG_24BPP  0x0200
#define ZGGEIG_FRONT2 0x0100
#define ZGGEIG_SEITE2 0x0400
#define ZGGEIG_HSTR   0x0800
#define ZGGEIG_HSTL   0x1000

/* Commands and options are case insensitive */

#define COMMAND_EXTRACT        'X'
#define COMMAND_COMBINEFZGSTD  'V'
#define COMMAND_COMBINEFZGUSER 'W'
#define COMMAND_COMBINEGZG     'S'
#define COMMAND_COMBINEUZG     'U'
#define COMMAND_ZOOM1          '1'
#define COMMAND_ZOOM2          '2'
#define COMMAND_ZOOM4          '4'

#define OPTION_OVERWRITE   'O'
#define OPTION_SIMULATE    'S'
#define OPTION_INFO        'I'

 /* Maximum count of vehicles in one vehicle set */
#define FZGF_MAXFZGZAHL           (16)
 /* Maximum count of vehicle sets in one user-defined file */
#define MAXNUTZERFZGSATZ_PRODATEI (10)

typedef struct FZGELEM * PFZGELEM;
typedef struct FZGELEM {
                  PFZGELEM pNext; /* Pointer to next, =NULL if last one */
                  INT32    FzgSatzNr; /* Number of vehicle set */
                  INT32    FzgLokNr;  /* Number of vehicle in the set */
		  INT32    Len;   /* Length of single file in BYTE */
                  INT32    Zeit;  /* Date+Time, see there */
                  CHAR8    PfName[MAXPATH]; /* Path+Name */
	       } FZGELEM;

typedef struct GZGELEM * PGZGELEM;
typedef struct GZGELEM {
                  PGZGELEM pNext; /* Pointer to next, =NULL if last one */
                  INT32    SymNr; /* Number, Standard: global
                                             User-defined: local in the file */
                  INT32    Bauform; /* Bauform/Version
                                       (for std. driving way, else = 0) */
                  INT32    FwSig;   /* HIWORD: Fahrweg/Driving way 1..4
                                       (for std. driving way, else = 0)
                                       FAHRWEGTYP_xxx + 1
                                       LOWORD: Signal Phase = 1 or 2
                                       (for blinking signals
                                        i.e. grade crossings
                                        else (mostly) = 0) */
                  INT32    Phase;   /* Animation Phase
                                       (for driving way = 0) */
                  INT32    Alt;     /* Alternative (0 or 1..4) */
		  INT32    Len;   /* Length of file in BYTE */
                  INT32    Zeit;  /* Date+Time, see there */
                  CHAR8    PfName[MAXPATH]; /* Path+Name (not very efficient...) */
	       } GZGELEM;

/* Count of stored file informations */
STATIC INT32  Anzahl_Stored = 0;

/* Counter for some statistical information
   (for use of "Info" option) */
STATIC INT32  Anzahl_F2, Anzahl_S2, Anzahl_H2, Anzahl_HR, Anzahl_HL;   

STATIC INT32  Sprach_Code;

typedef struct TXTELEM * PTXTELEM;
typedef struct TXTELEM CONST * PCTXTELEM;
typedef struct TXTELEM {
                  CONST INT32 TxtNr; /* Number of text */
                  /* Pointer to Text: SPRACHE_MIN.._MAX
                                      _ENG,_DEU,_POL,_ESP,_CZE */
                  CONST PCCHAR8 pTxt[ SPRACHE_ANZAHL ];
	       } TXTELEM;

#define TXTNR_UNDEF  -1
#define TXTNR_000     0
#define TXTNR_001     1
#define TXTNR_002     2
#define TXTNR_003     3
#define TXTNR_004     4
#define TXTNR_005     5
#define TXTNR_006     6
#define TXTNR_007     7
#define TXTNR_008     8
#define TXTNR_009     9
#define TXTNR_010    10
#define TXTNR_011    11
#define TXTNR_012    12
#define TXTNR_013    13
#define TXTNR_014    14
#define TXTNR_015    15
#define TXTNR_016    16
#define TXTNR_017    17
#define TXTNR_F_LESEN  18
#define TXTNR_F_DATEN  19
#define TXTNR_COLLECT  20
#define TXTNR_COPY     21
#define TXTNR_FRONT_2  22
#define TXTNR_SEITE_2  23
#define TXTNR_HST_2RL  24
#define TXTNR_LEN      25
#define TXTNR_HOS      26
#define TXTNR_BRF      27
#define TXTNR_HOF      28
#define TXTNR_F_S_T    29
#define TXTNR_F_S_T_Z  30 

#define MIN_FAHRWEG             0x0000
#define MAX_FAHRWEG             0x3FFF

/* Length of driving way blocks of equal function */
#define SYMDIFF_GLBLK           0x0200
#define LD_SYMDIFF_GLBLK        9  /* logarithmus dualis */
#if ((1 << LD_SYMDIFF_GLBLK ) != SYMDIFF_GLBLK )
#error Update!
#endif

#define USERGSYMZAHL_PER_DATEI  90  /* until B384: (2*36) */
/* smallest and largest scenic element
   incl. user-defined */
#define MIN_LSYMBOL      0xC000
#define MAX_LSYMBOL      0xFFFF
/* Animationsphasen von Landschaftssymbolen */
#define MIN_ANIPHASE     0
#define MAX_ANIPHASE    99
/* Alternativ-Varianten von Landschaftssymbolen */
#define KEINE_ALTERNATIVE 0  /* = alle Alternativen wenn mehrere ex. */
#define MIN_ALTERNATIVE   1
#define MAX_ALTERNATIVE   4
#define MAX_ALTERNATIVZAHL ( MAX_ALTERNATIVE - MIN_ALTERNATIVE + 1 )

/* Fahrwegtypen / Driving way types */
#define ANZAHL_FAHRWEGTYPEN  4
#define FAHRWEGTYP_GLSTR     0
#define FAHRWEGTYP_GLBBK     1
#define FAHRWEGTYP_H2O       2
#define FAHRWEGTYP_BUS       3
CONST CHAR Fahrwegzeichen[ ANZAHL_FAHRWEGTYPEN ] = {'s','t','h','b'};
#define ZEICHEN_FWMASKE   'm'  /* driving way mask */

#define FILEBUFSIZE 4096
STATIC BYTE   WriteBuf[ FILEBUFSIZE ]; /* Dateipuffer (Schreiben) */
STATIC INT32  WritePos = 0;      /* aktuelle Position beim Schreiben */
STATIC BYTE   ReadBuf[ FILEBUFSIZE ]; /* Dateipuffer (Lesen) */
STATIC INT32  ReadPos = 0;      /* aktuelle Position beim Lesen */

STATIC PFZGELEM p_FzgElemente = NULL;
STATIC PGZGELEM p_GzgElemente = NULL;

/* lokale Fktn. */

STATIC VOID FASTCALL InfoUserEnd( VOID );
STATIC VOID FASTCALL InitAll( VOID );
STATIC VOID FASTCALL QuitAll( VOID );
STATIC INT32 FASTCALL EinpackenFzg( PCCHAR8 pFileName, PCCHAR8 pDirectory,
                                    INT32 Zoomfaktor, BOOL Nutzerdef,
                                    BOOL Simulieren, BOOL Info,
                                    PCHAR pErrFile );
STATIC INT32 FASTCALL EinpackenGzg( PCHAR8 pFileName, PCHAR8 pDirectory,
                                    INT32 Zoomfaktor, BOOL Nutzerdef,
                                    BOOL Simulieren, BOOL Info,
                                    PCHAR pErrFile );
STATIC INT32 FASTCALL Auspacken( PCCHAR8 pFileName, PCCHAR8 pDirectory,
                                 BOOL Ueberschreiben, BOOL Simulieren,
                                 PINT32 pFound,
                                 PINT32 pCopied, PINT32 pSkipped );
STATIC INT32 FASTCALL WriteSeekTableFzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL WriteDatenFzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL StoreNameFzg( PCCHAR8 pPfad, PCCHAR8 pName,
                                    INT32 len, INT32 datum, INT32 zeit,
                                    BOOL Nutzerdef, BOOL Info );
STATIC INT32 FASTCALL WriteSeekTableGzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL WriteDatenGzg( FILE * file2, BOOL Simulieren );
STATIC INT32 FASTCALL StoreNameGzg( PCHAR8 pPfad, PCHAR8 pName,
                                    INT32 Len, INT32 datum, INT32 zeit,
                                    BOOL Nutzerdef, PCHAR Archivname,
                                    BOOL Info );
STATIC VOID FASTCALL Fehler( PCCHAR8 pText );
STATIC VOID FASTCALL Trace( PCCHAR8 pText );
STATIC VOID FASTCALL Meldung( PCCHAR8 pText, PCCHAR8 pTitel );
STATIC CHAR FASTCALL Caps( CHAR zei );
STATIC VOID FASTCALL CapsString( PCHAR8 Text );
STATIC BOOL FASTCALL DateiGibtEs( PCCHAR8 pName );
STATIC FILE * FASTCALL OpenWrite( PCCHAR8 pName );
STATIC INT32 FASTCALL WriteByte( FILE * file2, BYTE b );
STATIC INT32 FASTCALL WriteWord16( FILE * file2, WORD16 w );
STATIC INT32 FASTCALL WriteInt32( FILE * file2, INT32 i );
STATIC INT32 FASTCALL CloseWrite( FILE * file2 );
STATIC FILE * FASTCALL OpenRead( PCCHAR8 pName );
STATIC INT32 FASTCALL ReadByte( FILE * file, PBYTE pB );
STATIC INT32 FASTCALL ReadWord16( FILE * file, PWORD16 pW );
STATIC INT32 FASTCALL ReadInt32( FILE * file, PINT32 pI );
STATIC INT32 FASTCALL ReadVorspann( FILE * file );
STATIC INT32 FASTCALL CloseRead( FILE * file );
STATIC PVOID FASTCALL AllocateMemory( INT32 len );
STATIC VOID FASTCALL FreeMemory( PVOID pDat );
STATIC INT32 FASTCALL ReadOverSpacePos( PCCHAR8 pBuf );
STATIC INT32 FASTCALL DelSpaces( PCHAR8 pBuf );
STATIC INT32 FASTCALL SetDateiAttr( PCCHAR8 pPfad, INT32 Zeit );
STATIC INT32 FASTCALL GetSprache( VOID );
STATIC PCCHAR8 FASTCALL GetText( INT32 Textnummer );
STATIC VOID FASTCALL MemMove1( PVOID pDest, PCVOID pSrc, INT32 len );
#define MemCopy1( pD, pS, nL ) MemMove1( pD, pS, nL )
STATIC INT32 FASTCALL GetGzgDateiName( INT32 GSym, INT32 Bauform, INT32 FwSig,
                                       INT32 AddInfo, INT32 Phase,
                                       PCHAR8 pFName, INT32 Zf,
                                       PCCHAR8 pNutzerName );
STATIC VOID FASTCALL stradd( PCHAR8 pStr, CHAR8 Zei );
STATIC VOID FASTCALL SubExt( PCHAR8 pName );
STATIC VOID FASTCALL printfNL( VOID );

/*--------------------------------------------------------------*/
/* MAIN */

#pragma argsused
INT32 main( INT32 argc, PCHAR8 argv[] )
  {
   CHAR8  ArcFileName[ MAXPATH ];
   CHAR8  DirectoryName[ MAXPATH ];
   CHAR8  ErrFileName[ MAXPATH ];
   FILE   * file1;
   INT32  rc, optbuflen, optzei0, optzei1, zoomfaktor, len;
   INT32  anzfound, anzcopy, anzskip;
   CHAR8  errbuf[MAXPATH+120], optbuf[20];
   CHAR8  extbuf[MAXPATH+120+1000];
   PCHAR8 pOptBuf;
   BOOL   auspacken, fahrzeuge, nutzerdef, overwrite, simulate, info;

     /* Sprache ermitteln / Get Language */
     Sprach_Code = GetSprache();

     printfNL();
     printf( "JB Data-Combine %s 0.023 02/2011", GetText( TXTNR_001 ) );
     printfNL();

     if ( _argc < 4 )
        /* 0 is the .exe file self, 1..4 are parameters */
       {
        InfoUserEnd();
        return 1;
       }

     optbuflen = strlen( _argv[ 1 ] ) + 1;
     MemCopy1( optbuf, _argv[ 1 ], min( optbuflen, sizeof( optbuf ) ) );
     optbuf[ sizeof( optbuf ) - 2 ] = '\0';

     DelSpaces( optbuf );
     pOptBuf = optbuf;

     if (( *pOptBuf == '-' ) || ( *pOptBuf == '/' ))
       {
        pOptBuf++;
        DelSpaces( pOptBuf );
       }

     auspacken = fahrzeuge = nutzerdef = FALSE;

     optzei0 = Caps( *pOptBuf++ );
     if ( optzei0 == COMMAND_EXTRACT )
        auspacken = TRUE;
     else
       {
        if ( optzei0 == COMMAND_COMBINEFZGSTD )
           fahrzeuge = TRUE;
        else
          {
           if ( optzei0 == COMMAND_COMBINEFZGUSER )
              fahrzeuge = nutzerdef = TRUE;
           else
             {
              if ( optzei0 == COMMAND_COMBINEUZG )
                 nutzerdef = TRUE;
              else
                {
                 if ( optzei0 != COMMAND_COMBINEGZG )
                   {
                    InfoUserEnd();
                    return 1;
                   }
                }
             }
          }   
       }

     if ( !auspacken )
       {
        DelSpaces( pOptBuf );
        optzei1 = Caps( *pOptBuf++ );
        if ( optzei1 == COMMAND_ZOOM1 )
           zoomfaktor = 1;
        else
          {
           if ( optzei1 == COMMAND_ZOOM2 )
              zoomfaktor = 2;
           else
             {
              if ( optzei1 == COMMAND_ZOOM4 )
                 zoomfaktor = 4;
              else
                {
                 InfoUserEnd();
                 return 1;
                }
             }
          }
       }
     else /* Extract */
       {
        pOptBuf++;
       }

     /* Argument 2: Directory */
     strcpy( DirectoryName, _argv[ 2 ] );
     len = strlen( DirectoryName );
     if ( len > 0 )
       {
        if ( DirectoryName[ len - 1 ] == '\\' )
           DirectoryName[ len - 1 ] = '\0';
       }

     /* Argument 3: Filename */
     strcpy( ArcFileName, _argv[ 3 ] );
     if ( !auspacken )
       {
        if ( strchr( ArcFileName, '.' ) == NULL )
          {
           if ( fahrzeuge )
             {
              strcat( ArcFileName, FEXT_ARCZOOMFZG );
             }
           else
             {
              if ( nutzerdef )
                {
                 strcat( ArcFileName, FEXT_ARCZOOMUZG );
                }
              else
                {
                 strcat( ArcFileName, FEXT_ARCZOOMGZG );
                }
             }
           stradd( ArcFileName, LOBYTE( '0'+zoomfaktor ) );
          }
       }   

     /* Argument 4: More options */
     overwrite = simulate = info = FALSE;
     if ( _argc >= 5 )
       {
        optbuflen = strlen( _argv[ 4 ] ) + 1;
        MemCopy1( optbuf, _argv[ 4 ], min( optbuflen, sizeof( optbuf ) ) );

        optbuf[ sizeof( optbuf ) - 2 ] = '\0';

        DelSpaces( optbuf );
        pOptBuf = optbuf;

        if (( *pOptBuf == '-' ) || ( *pOptBuf == '/' ))
          {
           pOptBuf++;
           DelSpaces( pOptBuf );
          }

        while ( *pOptBuf != '\0' )
          {
           optzei0 = Caps( *pOptBuf++ );
           if ( optzei0 == OPTION_OVERWRITE )
              overwrite = TRUE;
           else
              if ( optzei0 == OPTION_SIMULATE )
                 simulate = TRUE;
              else
                 if ( optzei0 == OPTION_INFO )
                    info = TRUE;
          }
       }

     file1 = OpenRead( ArcFileName );
     CloseRead( file1 );

     if (file1 == NULL)
       {
        if ( auspacken )
          {
           sprintf( errbuf, GetText( TXTNR_002 ), ArcFileName );
           Fehler( errbuf );
           exit(1);
          }
       }
     else /* opened OK */
       {
        if ( !auspacken )
          {
           if ( !overwrite )
             {
              sprintf( errbuf, GetText( TXTNR_003 ), ArcFileName );
              Fehler( errbuf );
              exit(1);
             } 
          }
       }

     InitAll();

     if ( auspacken ) /* extract */
       {
        rc = Auspacken( ArcFileName, DirectoryName, overwrite, simulate,
                        &anzfound, &anzcopy, &anzskip );

        if ( rc == 0 )
          {
           sprintf( errbuf, GetText( TXTNR_007 ), anzfound, anzcopy, anzskip );
           Meldung( errbuf, GetText( TXTNR_008 ) );
          }
        else
          {
           sprintf( errbuf, "%s%d",
                            GetText( TXTNR_009 ),
                            rc );
           Fehler( errbuf );
          }
       }
     else /* combine */
       {
        if ( fahrzeuge )  /* vehicles */
          {
           rc = EinpackenFzg( ArcFileName, DirectoryName, zoomfaktor,
                              nutzerdef, simulate, info, ErrFileName );
          }
        else  /* scenery/way */
          {
           rc = EinpackenGzg( ArcFileName, DirectoryName, zoomfaktor,
                              nutzerdef, simulate, info, ErrFileName );
          }

        switch ( rc )
          {
           case 0: /* OK */
              sprintf( errbuf, GetText( TXTNR_010 ),
                       Anzahl_Stored );

              if (( info ) && ( fahrzeuge ))
                {
                 sprintf( extbuf, "%s\n\n %s=%d %s=%d %s=%d,%d,%d",
                          errbuf,
                          GetText( TXTNR_FRONT_2 ), Anzahl_F2,
                          GetText( TXTNR_SEITE_2 ), Anzahl_S2,
                          GetText( TXTNR_HST_2RL ),
                             Anzahl_H2, Anzahl_HR, Anzahl_HL );
                 Meldung( extbuf, GetText( TXTNR_008 ) );
                }
              else
                 Meldung( errbuf, GetText( TXTNR_008 ) );

              break;

           case 1: /* No file found */
              Fehler( GetText( TXTNR_014 ) );
              break;

           case 3: /* No archive created */
              Fehler( GetText( TXTNR_016 ) );
              break;

           case 4: /* Write error */
              Fehler( GetText( TXTNR_017 ) );
              break;

           case 11: /* Open/Read error */
              sprintf( errbuf, "%s: \"%s\"",
                                GetText( TXTNR_F_LESEN ), ErrFileName );
              Fehler( errbuf );
              break;

           case 12:  /* out of mem */
              sprintf( errbuf, "%s: \"%s\"",
                                GetText( TXTNR_015 ), ErrFileName );
              Fehler( errbuf );
              break;

           case 13: /* Data error */
              sprintf( errbuf, "%s: \"%s\"",
                                GetText( TXTNR_F_DATEN ), ErrFileName );
              Fehler( errbuf );
              break;

           default: /* Something unknown: telling the rc */
              sprintf( errbuf, "%s%i",
                                GetText( TXTNR_011 ), rc );
              Fehler( errbuf );
          }
       }

     return 0;
    }

/* --------------------------------------------------------------------- */

VOID FASTCALL InitAll( VOID )
    /* Initialisierungen */
    {
     p_FzgElemente = NULL;
     p_GzgElemente = NULL;
     Anzahl_Stored = 0;
     Anzahl_F2 = Anzahl_S2 = Anzahl_H2 = Anzahl_HR = Anzahl_HL = 0;
    }

/* --------------------------------------------------------------------- */

VOID FASTCALL QuitAll( VOID )
    /* Freigaben */
  {
   PFZGELEM pT1, pT2;
   PGZGELEM pG1, pG2;

     pT1 = p_FzgElemente;
     p_FzgElemente = NULL;
     while ( pT1 != NULL )
       {
        pT2 = pT1->pNext;
        FreeMemory( pT1 );
        pT1 = pT2;
       }

     pG1 = p_GzgElemente;
     p_GzgElemente = NULL;
     while ( pG1 != NULL )
       {
        pG2 = pG1->pNext;
        FreeMemory( pG1 );
        pG1 = pG2;
       }
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL EinpackenFzg( PCCHAR8 pFileName, PCCHAR8 pDirectory,
                                    INT32 Zoomfaktor, BOOL Nutzerdef,
                                    BOOL Simulieren,  BOOL Info,
                                    PCHAR8 pErrFile )
 /* Creates an archive file, i.e. combines data to pack vehicles
    IN:  pFilename = Archive file name (file may get overwritten)
                     can include a path (is added to Directory)
         pDirectory = Source path
         ZoomFaktor = 2 or 4
         Nutzerdef  = FALSE: Create standard archive file from single files
                             with filenames "nnnnn-nn.fzn"
                      TRUE: Create user-defined archive file from single files
                            with filenames "FileName-nn-nn.fzn"
         Simulieren = TRUE: write nothing (simulate only)
         Info       = TRUE: output of some statistics data to stdout
         pErrFile   = Buffer for filename of single file with problem
                      CHAR8[MAXPATH]
    OUT: 0 OK
         1 No file found
         3 No archive file created
         4 Error writing archive file
         5 Wrong parameter
        11 Error while Open/Read
        12 Out of Memory
        13 Data error
        for 11..13: pErrFile contains filename of related single file */
  {
   CHAR8 Pfad[ MAXPATH ], PfadMask[ MAXPATH ];
   struct ffblk fblk;
   FILE * file2;
   INT32  rc, i;
   CONST CHAR8 Vsp[] = "JB BAHN Vehicle gfx Zoom";
   PCCHAR8 pZei;

     if (( Zoomfaktor != 2 ) && ( Zoomfaktor != 4 ))
        return 5;

     if (( pFileName == NULL ) || ( pDirectory == NULL ))
        return 5;

     if ( strlen( pFileName ) + 60 >= sizeof( Pfad ) )
        return 5;

     if ( pErrFile == NULL )
        return 5;

     strcpy( Pfad, pDirectory );

     strcpy( PfadMask, Pfad );

     if ( Nutzerdef )
       {
        stradd( PfadMask, '\\' );
        strcat( PfadMask, pFileName );
        SubExt( PfadMask );
        strcat( PfadMask, "-??-??." );
       }
     else /* Standard file */
       {
        strcat( PfadMask, "\\?????-??." );
       }

     if ( Zoomfaktor == 2 )
        strcat( PfadMask, FEXT_ZOOM2FZG );
     else
        strcat( PfadMask, FEXT_ZOOM4FZG );

     /* Remark: These masks accept any character for '?',
        but we are interested in decimal digits only.
        We could filter it in detail here, but for what? */

     if ( findfirst( PfadMask, &fblk, 0 ) != 0 )
        return 1;

     if ( Info )
       {
        printf( GetText( TXTNR_COLLECT ) );
        printfNL();
       }

     rc = StoreNameFzg( Pfad, fblk.ff_name,
                        fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime,
                        Nutzerdef, Info );
     if ( rc )
    /* 1 Error while Open/Read
       2 Out of Memory
       3 Data error */
       {
        strcpy( pErrFile, fblk.ff_name );
        return rc+10;
       }

     while ( findnext( &fblk ) == 0 )
       {
        rc = StoreNameFzg( Pfad, fblk.ff_name,
                           fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime,
                           Nutzerdef, Info );
        if ( rc )
          {
           strcpy( pErrFile, fblk.ff_name );
           return rc+10;
          } 
       }

     if ( ! Simulieren )
       {
        file2 = OpenWrite( pFileName );
        if ( file2 == NULL )
           return 3;

        /* Write text header */
        pZei = Vsp;
        while ( *pZei != '\0' )
          {
           rc = WriteByte( file2, *pZei++ );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        rc = WriteByte( file2, CTRL_Z );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
        rc = WriteByte( file2, '\0' );  /* added with 3.86 */
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        /* Write Identifier */
        for ( i = 0; i < sizeof( IDENT_CODE_FZX ); i++ )
          {
           rc = WriteByte( file2, IDENT_CODE_FZX[i] );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        /* Archive file version, added with 3.86 */
        rc = WriteInt32( file2, ZOOMXARCHIVEFZGFORMAT_386_0 );  
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
       }

     /* write seek table */
     rc = WriteSeekTableFzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     /* copy single files */
     rc = WriteDatenFzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     if ( ! Simulieren )
       {
        if ( CloseWrite( file2 ) )
           return 4;
       }

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL EinpackenGzg( PCHAR FileName, PCHAR Directory,
                                    INT32 Zoomfaktor, BOOL Nutzerdef,
                                    BOOL Simulieren, BOOL Info,
                                    PCHAR8 pErrFile )
  /* Stores/packs data (scenery)
       IN:  Filename = Archive file (will be overwritten when already ex.)
                       can include a path
            Directory = Source path
            ZoomFaktor = 1, 2 or 4
            Nutzerdef  = TRUE: user-defined data
            Simulieren = TRUE: do not write but simulate only
            Info       = TRUE: output some information
            pErrFile   = Buffer for filename of single file with problem
                         CHAR8[MAXPATH]
       OUT: 0 OK
            1 No file found
            3 No archive file created
            4 Error writing archive file
            5 Wrong parameter
           11 Error while Open/Read
           12 Out of Memory
           13 Data error
           for 11..13: pErrFile contains filename of single file */
  {
   CHAR8 Pfad[ MAXPATH ], PfadMask[ MAXPATH ], Dateiname[ MAXPATH ];
   struct ffblk fblk;
   FILE * file2;
   INT32  rc, i;
   STATIC CHAR Vsp1[] = "JB BAHN Scenery gfx Zoom";
   STATIC CHAR Vsp2[] = "\r\ncreated by FZGC2";
   PCCHAR8 pZei;
   PCCHAR8 pIDC;

     if (( Zoomfaktor != 1 ) && ( Zoomfaktor != 2 ) && ( Zoomfaktor != 4 ))
        return 5;

     if ( strlen( FileName ) + 60 >= sizeof( Pfad ) )
        return 5;

     if ( pErrFile == NULL )
        return 5;

     strcpy( Pfad, Directory );
     strcpy( PfadMask, Pfad );
     stradd( PfadMask, '\\' );

     if ( Nutzerdef )
       {
        rc = fnsplit( FileName, NULL, NULL, Dateiname, NULL );
        if (( rc & WILDCARDS ) || !( rc & FILENAME ))
           return 5;
        stradd( PfadMask, '\\' );
        strcat( PfadMask, Dateiname );
       }
     strcat( PfadMask, "*." );

     if ( Zoomfaktor == 2 )
        strcat( PfadMask, FEXT_ZOOM2GZG );
     else
       {
        if ( Zoomfaktor == 4 )
           strcat( PfadMask, FEXT_ZOOM4GZG );
        else
           strcat( PfadMask, FEXT_ZOOM1GZG );
       }

     /* look for files */

     if ( findfirst( PfadMask, &fblk, 0 ) != 0 )
       {
        Fehler( GetText( TXTNR_014 ) );
        return 1;
       }

     if ( Info )
       {
        printf( GetText( TXTNR_COLLECT ) );
        printfNL();
       }

     rc = StoreNameGzg( Pfad, fblk.ff_name,
                        fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime,
                        Nutzerdef, Dateiname, Info );
     if ( rc )
       {
        strcpy( pErrFile, fblk.ff_name );
        return 10+rc;
       }

     while ( findnext( &fblk ) == 0 )
       {
        rc = StoreNameGzg( Pfad, fblk.ff_name,
                           fblk.ff_fsize, fblk.ff_fdate, fblk.ff_ftime,
                           Nutzerdef, Dateiname, Info );
        if ( rc )
          {
           strcpy( pErrFile, fblk.ff_name );
           return 10+rc;
          }
       }

     /* write collected data */

     if ( ! Simulieren )
       {
        file2 = OpenWrite( FileName );
        if ( file2 == NULL )
           return 3;

        /* write header */
        pZei = Vsp1;
        while ( *pZei != '\0' )
          {
           rc = WriteByte( file2, *pZei++ );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
        pZei = Vsp2;
        while ( *pZei != '\0' )
          {
           rc = WriteByte( file2, *pZei++ );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }

        rc = WriteByte( file2, CTRL_Z );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }

        /* Kennung schreiben */
        if ( Nutzerdef )
           pIDC = IDENT_CODE_UZX;
        else
           pIDC = IDENT_CODE_GZX;
        for ( i = 0; i < sizeof( IDENT_CODE_GZX ); i++ )
          {
           rc = WriteByte( file2, pIDC[i] );
           if ( rc )
             {
              CloseWrite( file2 );
              return 4;
             }
          }
        rc = WriteByte( file2, LOBYTE( '0'+Zoomfaktor ) );
        if ( rc )
          {
           CloseWrite( file2 );
           return 4;
          }
       }

     /* Write seek table */
     rc = WriteSeekTableGzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     /* Copy single files */
     rc = WriteDatenGzg( file2, Simulieren );
     if ( rc )
       {
        CloseWrite( file2 );
        return 4;
       }

     if ( ! Simulieren )
       {
        if ( CloseWrite( file2 ) )
           return 4;
       }

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL StoreNameFzg( PCCHAR8 pPfad, PCCHAR8 pName,
                                    INT32 len, INT32 datum, INT32 zeit,
                                    BOOL Nutzerdef, BOOL Info )
/* Stores name of single file and needed properties
   (for vehicles)
   IN: pPfad = Path
       pName = Filename of the single file
       len   = File length in BYTE
       datum = Bit0..4 day, Bit5..8 month, Bit9..15 year since 1980
       zeit  = Bit0..4 sec/2, Bit5..10 min, Bit11..15 hour
       Info  = TRUE: store and output some statistics information
       Nutzerdef = TRUE:  Handle filenames as user-defined
                   FALSE: Handle filenames as for standard vehicles
  OUT: 0 OK
       1 Error while Open/Read
       2 Out of Memory
       3 Data error */
  {
   INT32    satznr, lokfzgnr, rc, pos;
   PFZGELEM pEl, pV, pN;
   FILE *   file1;
   WORD16   w16;
   BYTE     bl, bh;
   INT32    fversion, fsubversion, idcode,
            fzgeig, fzglen, fzghow, fzgbr, fzghos;
   CHAR8    frontzei, seitezei, hstzei;

     if ( len <= 0 )
        return 1;

     if (( pPfad == NULL ) || ( pName == NULL ))
        return 3;

     if ( Nutzerdef )
       {
        /* look for "-nn-nn" in the filename backwards from end */
        pos = strlen( pName ) - 1;
        if ( pos <= 0 )
           return 3;

        while ( pName[pos--] != '.' )
          {
           if ( pos < 0 )
              return 3;
          }
        while ( pName[pos--] != '-' )
          {
           if ( pos < 0 )
              return 3;
          }
        while ( pName[pos--] != '-' )
          {
           if ( pos < 0 )
              return 3;
          }
        pos++;  

        if ( sscanf( &pName[pos], "-%d-%d", &satznr, &lokfzgnr ) != 2 )
           return 3;
        if ( satznr >= MAXNUTZERFZGSATZ_PRODATEI )
           return 3;
       }
     else
       {
        if ( sscanf( pName, "%d-%d", &satznr, &lokfzgnr ) != 2 )
           return 3;
       }

     if (( satznr < 0 ) || ( lokfzgnr < 0 ) || ( lokfzgnr > FZGF_MAXFZGZAHL ))
        return 3;

     if ( strlen( pPfad ) + strlen( pName ) >= MAXPATH-1 )
        return 3;

     pEl = AllocateMemory( sizeof( FZGELEM ) );
     if ( pEl == NULL )
        return 2;

     pEl->FzgSatzNr = satznr;
     pEl->FzgLokNr  = lokfzgnr;
     pEl->Len       = len;
     pEl->Zeit      = (datum << 16) | zeit;
     strcpy( pEl->PfName, pPfad );
     strcat( pEl->PfName, "\\" );
     strcat( pEl->PfName, pName );

     /* Info if requested */
     if ( Info )
       {
        printf( pEl->PfName );
        printfNL();

        file1 = OpenRead( pEl->PfName );
        if ( file1 == NULL )
           return 1;

        rc = ReadVorspann( file1 );
        if ( rc )
           return 1;

        rc = ReadInt32( file1, &idcode );
        if ( rc )
           return 1;
        /* idcode could be checked here for "FZG2" or "FZG4" */

        rc = ReadByte( file1, &bh );
        if ( rc )
           return 1;
        rc = ReadByte( file1, &bl );
        if ( rc )
           return 1;
        fversion = ( bh << 8 ) | bl;
        if ( fversion == ZOOM2FZGFORMAT383 )
          {
           rc = ReadByte( file1, &bl );
           if ( rc )
              return 1;
           fsubversion = bl;
          }
        else
          {
           if ( fversion == ZOOMXFZGFORMAT385 )
             {
              rc = ReadByte( file1, &bl );
              if ( rc )
                 return 1;
              rc = ReadByte( file1, &bh );
              if ( rc )
                 return 1;
              fsubversion = ( bh << 8 ) | bl;

              rc = ReadInt32( file1, &fzgeig );
              if ( rc )
                 return 1;
             }
           else
             {
              CloseRead( file1 );
              return 3;
             }
          }

        printf( "%04x %04x", fversion, fsubversion );

        rc = ReadWord16( file1, &w16 );
        if ( rc )
           return 1;
        fzglen = w16;
        rc = ReadWord16( file1, &w16 );
        if ( rc )
           return 1;
        fzghow = w16;
        rc = ReadWord16( file1, &w16 );
        if ( rc )
           return 1;
        fzgbr = w16;
        rc = ReadWord16( file1, &w16 );
        if ( rc )
           return 1;
        fzghos = w16;
        if ( fversion == ZOOM2FZGFORMAT383 )
          {
           rc = ReadWord16( file1, &w16 );
           if ( rc )
              return 1;
           fzgeig = w16;   
          }

        CloseRead( file1 );

        printf( " %s=%d %s=%d %s=%d %s=%d",
                GetText( TXTNR_LEN ), fzglen,
                GetText( TXTNR_HOS ), fzghow,
                GetText( TXTNR_BRF ), fzgbr,
                GetText( TXTNR_HOF  ), fzghos );

        if ( fzgeig & ZGGEIG_FRONT2 )
          {
           frontzei = GetText( TXTNR_F_S_T_Z )[1];
           Anzahl_F2++;
          }
        else
          {
           frontzei = GetText( TXTNR_F_S_T_Z )[0];
          }
        if ( fzgeig & ZGGEIG_SEITE2 )
          {
           seitezei = GetText( TXTNR_F_S_T_Z )[3];
           Anzahl_S2++;
          }
        else
          {
           seitezei = GetText( TXTNR_F_S_T_Z )[2];
          }
        if ( fzgeig & ZGGEIG_HSTR )
          {
           if ( fzgeig & ZGGEIG_HSTL )
             {
              hstzei = GetText( TXTNR_F_S_T_Z )[7];
              Anzahl_H2++;
             }
           else
             {
              hstzei = GetText( TXTNR_F_S_T_Z )[5];
              Anzahl_HR++;
             }
          }
        else
          {
           if ( fzgeig & ZGGEIG_HSTL )
             {
              hstzei = GetText( TXTNR_F_S_T_Z )[6];
              Anzahl_HL++;
             }
           else
             {
              hstzei = GetText( TXTNR_F_S_T_Z )[4];
             }
          }

        printf( GetText( TXTNR_F_S_T ),
                frontzei, seitezei, hstzei );
        printfNL();
       }

     /* InQueue, sorted */
     pV = NULL;
     pN = p_FzgElemente;

     while ( pN != NULL )
       {
        if ( pN->FzgSatzNr > satznr )
           break;
        if (( pN->FzgSatzNr == satznr ) && ( pN->FzgLokNr > lokfzgnr ))
           break;

        pV = pN;
        pN = pN->pNext;
       }

     if ( pV == NULL )
       {
        pEl->pNext = p_FzgElemente;
        p_FzgElemente = pEl;
       }
     else
       {
        pEl->pNext = pV->pNext;
        pV->pNext = pEl;
       }

     Anzahl_Stored++;

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteSeekTableFzg( FILE * file2, BOOL Simulieren )
 /* Writes table of set+vehicle numbers and seek()-positions
    IN:  file2      = File handle of opened output file
         Simulieren = TRUE: do not really write (simulate only)
   OUT:  0 = OK
         1 = Error while writing (file closed) */
  {
   INT32    len;
   PFZGELEM pTab = p_FzgElemente;

     /* Length of the seek table self including terminator */
     len = Anzahl_Stored * (3 * sizeof(INT32)) + sizeof(INT32);

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->FzgSatzNr ))
              return 1;
           if ( WriteInt32( file2, pTab->FzgLokNr ))
              return 1;
          }

        /* subtract the length of own seek entry,
           because this has been already read */
        len -= 3 * sizeof(INT32);

        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, len ) )
              return 1;
          }

        /* Summary length, + the numbers before the file, own length + time
           (that is a bit redundant, but makes the handling easier) */
        len += pTab->Len + 4 * sizeof( INT32 );

        pTab = pTab->pNext;
       }

     /* Terminator sign */
     if ( ! Simulieren )
       {
        if ( WriteInt32( file2, -1 ) )
           return 1;
       }

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteDatenFzg( FILE * file2, BOOL Simulieren )
 /* Writes the vehicle data, i.e. copies the single files
    (for vehicles)
    IN:  file2 = Handle of opened output file
         Simulieren = TRUE: Do not really write (simulate only)
   OUT:  0 = OK
         1 = Error while writing
         2 = Error while reading
    Close the output file after calling this function! */
  {
   INT32    len, rc;
   PFZGELEM pTab = p_FzgElemente;
   FILE *   file1;
   BYTE     b;

     printf( GetText( TXTNR_COPY ) );
     printfNL();

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->FzgSatzNr ))
              return 1;
           if ( WriteInt32( file2, pTab->FzgLokNr ))
              return 1;
           if ( WriteInt32( file2, pTab->Len ))
              return 1;
           if ( WriteInt32( file2, pTab->Zeit ))
              return 1;
          }

        printf( pTab->PfName );
        printfNL();

        file1 = OpenRead( pTab->PfName );
        if ( file1 == NULL )
           return 2;

        len = pTab->Len;
        while ( len-- )
          {
           rc = ReadByte( file1, &b );
           if ( rc )
              return 2;
           if ( ! Simulieren )
             {
              rc = WriteByte( file2, b );
              if ( rc )
                {
                 CloseRead( file1 );
                 return 1;
                }
             }
          }
        CloseRead( file1 );

        pTab = pTab->pNext;
       }

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL StoreNameGzg( PCHAR pPfad, PCHAR pName,
                                    INT32 Len, INT32 datum, INT32 zeit,
                                    BOOL Nutzerdef, PCHAR Archivname,
                                    BOOL Info )
/* Stores name of single file and needed properties (scenery)
   IN: pPfad = Path
       pName = Filename
       Len   = Length in BYTE
       datum = Bit0..4 Day, Bit5..8 Month, Bit9..15 Year since 1980
       zeit  = Bit0..4 sec/2, Bit5..10 min, Bit11..15 h
       Nutzerdef = TRUE: userdefined file
       Archivname = Filename of archive without extension
                    (only needed if Nutzerdef == TRUE)
       Info  = Output of some information
  OUT: 0 OK
       1 Problem while Open/Read
       2 Out of memory
       3 Data error */
  {
   INT32    GSym, Bauform, sig, Phase, Alt, fw, fwnr, scz, i, altzei;
   PGZGELEM pEl, pV, pN;
   INT32    Pos;
   BOOL     fahrweg, fwmaske;

     if ( Len <= 0 )
        return 1;

     fahrweg = fwmaske = FALSE;
     fw = 0;

     if ( Archivname != NULL )
        /* for case-insensitive compare */
        strlwr( Archivname );

     if ( ! Nutzerdef )
       {
        if ( pName[0] == ZEICHEN_FWMASKE )
           fwmaske = TRUE;
        else
          {
           for ( i = 0; i < ANZAHL_FAHRWEGTYPEN; i++ )
             {
              if ( pName[0] == Fahrwegzeichen[i] )
                {
                 fw = i + 1;
                 fahrweg = TRUE;
                 break;
                }
             }
          }
        Pos = 1;
       }
     else /* user-defined */
       {
        Pos = strlen( Archivname ) + 1;
       }

     Phase = MIN_ANIPHASE;
     Alt   = KEINE_ALTERNATIVE;
     Bauform = sig = 0;

     if ( fahrweg )  /* driving way (additional scenic data, no mask) */
       {
        sig = '\0'; /* %c writes LOBYTE only? */
        scz = sscanf( &pName[Pos], "%d-%d%c", &GSym, &Bauform, &sig );
        if ( scz == 3 )
          {
           if (( sig >= 'a' ) && ( sig <= 'd' ))
             {
              /* 1..4
                 until BAHN 3.86b1: a and b only
                 since BAHN 3.86b2: a,b,c,d */
              sig = sig - 'a' + 1;
             }
           else
             {
              /* Remark: Since BAHN 3.86b2: 'a' resp. 1 is used by default
                 but ignored */
              sig = 0;
             }
          }
        else
           sig = 0;

        if (( GSym < MIN_FAHRWEG ) || ( GSym > MAX_FAHRWEG ))
           return 3;
       }
     else
       {
        if ( fwmaske ) /* driving way mask */
          {
           sig = 0;
           scz = sscanf( &pName[Pos], "%d-%d-%d", &fwnr, &GSym, &Bauform );

           if ( scz != 3 )
              return 3;

           if ( Bauform != 0 )
              return 3;

           if (( GSym < MIN_FAHRWEG ) || ( GSym >= SYMDIFF_GLBLK ))
              return 3;

           GSym = ( fwnr << LD_SYMDIFF_GLBLK ) | GSym;

           if (( GSym < MIN_FAHRWEG ) || ( GSym > MAX_FAHRWEG ))
              return 3;
          }
        else  /* scenery */
          {
           fw = sig = 0;

           altzei = '\0'; /* %c writes LOBYTE only? */
           scz = sscanf( &pName[Pos], "%d-%d%c", &GSym, &Phase, &altzei );
           if ( scz == 3 )
             {
              if (( altzei >= 'a' ) && ( altzei <= 'd' ))
                 Alt = altzei - 'a' + 1;
              else
                 Alt = KEINE_ALTERNATIVE;
             }
           else
             {
              Alt = KEINE_ALTERNATIVE;

              if ( scz == 1 )
                {
                 Phase = MIN_ANIPHASE;
                 /* then try again without phase */
                 altzei = '\0'; /* %c writes LOBYTE only? */
                 scz = sscanf( &pName[Pos], "%d%c", &GSym, &altzei );
                 if ( scz == 2 )
                   {
                    if (( altzei >= 'a' ) && ( altzei <= 'd' ))
                       Alt = altzei - 'a' + 1;
                    else
                       Alt = KEINE_ALTERNATIVE;
                   }
                }
             }

           if ( Nutzerdef ) /* user-defined */
             {
              if (( GSym < 0 ) || ( GSym >= USERGSYMZAHL_PER_DATEI ))
                 return 3;
             }
           else
             {
              if (( GSym < MIN_LSYMBOL ) || ( GSym > MAX_LSYMBOL ))
                 return 3;
             }
          }
       }

     if (( Phase < MIN_ANIPHASE ) || ( Phase > MAX_ANIPHASE ))
        return 3;

     if (( Alt < MIN_ALTERNATIVE ) || ( Alt > MAX_ALTERNATIVE ))
       {
        if ( Alt != KEINE_ALTERNATIVE )
           return 3;
       }

     pEl = AllocateMemory( sizeof( GZGELEM ) );
     if ( pEl == NULL )
        return 2;

     pEl->pNext   = NULL;
     pEl->SymNr   = GSym;
     pEl->Bauform = Bauform;
     pEl->FwSig   = (fw << 16) | sig;
     pEl->Phase   = Phase;
     pEl->Alt     = Alt;
     pEl->Len     = Len;
     pEl->Zeit  = (datum << 16) | zeit;
     strcpy( pEl->PfName, pPfad );
     strcat( pEl->PfName, "\\" );
     strcat( pEl->PfName, pName );

     if ( Info )
       {
        printf( pEl->PfName );
        printfNL();
       }

     /* InQueue, sorted */
     pV = NULL;
     pN = p_GzgElemente;

     while ( pN != NULL )
       {
        if ( pN->SymNr > GSym )
           break;

        if ( pN->SymNr == GSym )
          {
           if (( pN->Bauform > Bauform ) || ( pN->Phase > Phase ))
              break;

           if (( pN->Bauform == Bauform ) && ( pN->Phase == Phase ))
              if ( pN->Alt > Alt )
                 break;
          }
        pV = pN;
        pN = pN->pNext;
       }

     if ( pV == NULL )
       {
        pEl->pNext = p_GzgElemente;
        p_GzgElemente = pEl;
       }
     else
       {
        pEl->pNext = pV->pNext;
        pV->pNext = pEl;
       }

     Anzahl_Stored++;  /* count */

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteSeekTableGzg( FILE * file2, BOOL Simulieren )
 /* Schreibt Tabelle der FzgNummern und seek()-Positionen
    (Fahrzeuge)
    IN:  file2 = Handle Datei
         Simulieren = TRUE: nichts wirklich schreiben
   OUT:  0 = OK
         1 = Schreibfehler (Datei geschlossen) */
  {
   INT32    len;
   PGZGELEM pTab = p_GzgElemente;

     /* Length of seek table self incl. terminator */
     len = Anzahl_Stored * (5+1) * sizeof(INT32) + sizeof(INT32);

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->SymNr ))
              return 1;
           if ( WriteInt32( file2, pTab->Bauform ))
              return 1;
           if ( WriteInt32( file2, pTab->FwSig ))
              return 1;
           if ( WriteInt32( file2, pTab->Phase ))
              return 1;
           if ( WriteInt32( file2, pTab->Alt ))
              return 1;
          }

        /* excl. the own seek entry, because it was already read */
        len -= 6*(sizeof(INT32));

        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, len ) )
              return 1;
          }

        /* Summary length, + before the file: the data again (own length + time)
           (that is a bit redundant but more simple to handle) */
        len += pTab->Len + (5+2) * sizeof( INT32 );

        pTab = pTab->pNext;
       }

     /* terminator sign */
     if ( ! Simulieren )
       {
        if ( WriteInt32( file2, -1 ) )
           return 1;
       }    

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL WriteDatenGzg( FILE * file2, BOOL Simulieren )
 /* Writes dat, i.e. copies single files (scenery)
    IN:  file2 = destination file handle
         Simulieren = TRUE: do not write really (simulate only)
   OUT:  0 = OK
         1 = Write error
         2 = Read error
    Close file2 after, also in case of error! */
  {
   INT32    len, rc;
   PGZGELEM pTab = p_GzgElemente;
   FILE *   file1;
   BYTE     b;

     printf( GetText( TXTNR_COPY ) );
     printfNL();

     while ( pTab != NULL )
       {
        if ( ! Simulieren )
          {
           if ( WriteInt32( file2, pTab->SymNr ))
              return 1;
           if ( WriteInt32( file2, pTab->Bauform ))
              return 1;
           if ( WriteInt32( file2, pTab->FwSig ))
              return 1;
           if ( WriteInt32( file2, pTab->Phase ))
              return 1;
           if ( WriteInt32( file2, pTab->Alt ))
              return 1;
           if ( WriteInt32( file2, pTab->Len ))
              return 1;
           if ( WriteInt32( file2, pTab->Zeit ))
              return 1;
          }

        printf( pTab->PfName );
        printfNL();

        file1 = OpenRead( pTab->PfName );
        if ( file1 == NULL )
           return 2;

        len = pTab->Len;
        while ( len-- )
          {
           rc = ReadByte( file1, &b );
           if ( rc )
              return 2;
           if ( ! Simulieren )
             {
              rc = WriteByte( file2, b );
              if ( rc )
                {
                 CloseRead( file1 );
                 return 1;
                }
             }   
          }
        CloseRead( file1 );

        pTab = pTab->pNext;
       }

     return 0;
   }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL Auspacken( PCCHAR8 pFileName, PCCHAR8 pDirectory,
                                 BOOL Ueberschreiben, BOOL Simulieren,
                                 PINT32 pFound,
                                 PINT32 pCopied, PINT32 pSkipped )
 /* Extracts file (depending on archive file)
    IN:  pFilename = Archive file name, can include path
         pDirectory = Destination path
         Ueberschreiben = TRUE: Overwrite target files when already existing
         Simulieren = TRUE: nichts wirklich schreiben
    OUT: 0 = OK, then *pFound   = Total count of found files
                      *pCopied  = Count of copied files
                      *pSkipped = Count of skipped files
         1 = Archive file not found
         2 = Error while reading
         4 = Error while writing (data medium full?)
         5 = Wrong parameter */
  {
   FILE * file1, * file2;
   INT32  rc, fversion, fzgnr, fzgsatznr, fzgloknr;
   INT32  pos, cnt, len, Zeit, i, anzahl, anzahlkop, anzahlueb;
   INT32  symnr, bauform, fwsig, phase, alt, Zoomfaktor;
   BYTE   b;
   CHAR8  Pfad[ MAXPATH ], PfadKpl[ MAXPATH ], Name[ MAXPATH ];
   CHAR8  idbuf[ max( sizeof( IDENT_CODE_FZX ), sizeof( IDENT_CODE_GZX ) ) ];
   BOOL   Fahrzeuge, Nutzerdef, schreiben, b386vs;
   CHAR8  Nutzerdateiname[ MAXPATH ];

     if (( pFileName == NULL ) || ( pDirectory == NULL ))
        return 5;

     if ( strlen( pFileName ) + 60 >= sizeof( Pfad ) )
        return 5;

     rc = fnsplit( pFileName, NULL, NULL, Nutzerdateiname, NULL );
     if (( rc & WILDCARDS ) || !( rc & FILENAME ))
        return 5;

     anzahl = anzahlkop = anzahlueb = 0;
     if ( pFound != NULL )
        *pFound = 0;
     if ( pCopied != NULL )
        *pCopied = 0;
     if ( pSkipped != NULL )
        *pSkipped = 0;

     strcpy( Pfad, pDirectory );

     file1 = OpenRead( pFileName );
     if ( file1 == NULL )
        return 1;

     rc = ReadVorspann( file1 );
     if ( rc )
        return 2;

     /* Here, a NULL-Byte may follow
        But if not, then it should be a part of the following Id */
     rc = ReadByte( file1, &b );
     if ( rc )
        return 2;
     if ( b == '\0' )
       {
        b386vs = TRUE;
        rc = ReadByte( file1, &b );
        if ( rc )
           return 2;
       }
     else
        b386vs = FALSE;

     idbuf[0] = b;

     /* Read identifier (from Byte 1) */
#if ( sizeof( IDENT_CODE_FZX ) != sizeof( IDENT_CODE_GZX ) )
#error So nicht!
#endif
     for ( i = 1; i < sizeof( idbuf ); i++ )
       {
        rc = ReadByte( file1, &b );
        if ( rc )
           return 2;
        idbuf[i] = b;
       }

     Fahrzeuge = Nutzerdef = FALSE;

     if ( memcmp( idbuf, IDENT_CODE_FZX, sizeof( idbuf ) ) == 0 )
        Fahrzeuge = TRUE;
     else
       {
        if ( memcmp( idbuf, IDENT_CODE_UZX, sizeof( idbuf ) ) == 0 )
           Nutzerdef = TRUE;
        else
           if ( memcmp( idbuf, IDENT_CODE_GZX, sizeof( idbuf ) ) != 0 )
              return 2;
       }

     rc = ReadByte( file1, &b );
     if ( rc )
        return 2;
     Zoomfaktor = b - '0';

     /* Archive File Version (used for vehicles since BAHN 3.86b2) */
     if (( Fahrzeuge ) && ( b386vs ))
       {
        rc = ReadInt32( file1, &fversion );
        if ( rc )
           return rc;
        if ( fversion != ZOOMXARCHIVEFZGFORMAT_386_0 )
           return 2;
        Nutzerdef = TRUE; /* see remark below when reading the table */
       }
     else  /* older formats (never user-defined because not supported) */
       {
        fversion = ZOOMXARCHIVEFZGFORMAT_386_alt;
       }

     /* Read seek table
        (read and skip, because we do not search for something)
        i.e. we could seek directly to the first entry instead...
        but we get the total count of files here  */
     while ( TRUE )
       {
        if ( Fahrzeuge )
          {
           if ( fversion < ZOOMXARCHIVEFZGFORMAT_386_0 )
             {
              rc = ReadInt32( file1, &fzgnr );
              if ( rc )
                 return 2;
              if ( fzgnr == 0 )
                 break;
             }
           else /* fversion >= ZOOMXARCHIVEFZGFORMAT_386_0 */
             {
              rc = ReadInt32( file1, &fzgsatznr );
              if ( rc )
                 return 2;
              if ( fzgsatznr == -1 )
                 break;
              rc = ReadInt32( file1, &fzgloknr );
              if ( rc )
                 return 2;
              fzgnr = (fzgsatznr << 4) | fzgloknr;

              /* For vehicles, we do not know here whether standard or
                 user-defined Zoom archive (added with BAHN3.86Beta2)
                 That could be forced by an additional option or an alternative
                 extract command.
                 Also, a simple test for filename may be sufficient here
                 ("bahn.fz2" or "bahn.fz4").
                 The test is done here via the vehicle set numbers stored in
                 the table:
                  User-defined use 0..MAXNUTZERFZGSATZ_PRODATEI
                    and begin mostly at 0,
                  Standard begins at 5000, but this may change in future...
                 */
               if ( fzgsatznr > MAXNUTZERFZGSATZ_PRODATEI )
                  Nutzerdef = FALSE;
             }
          }
        else /* no vehicles */
          {
           rc = ReadInt32( file1, &symnr );
           if ( rc )
              return 2;
           if ( symnr == -1 )
              break;
           rc = ReadInt32( file1, &bauform );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &fwsig );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &phase );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &alt );
           if ( rc )
              return 2;
          }

        rc = ReadInt32( file1, &pos );
        if ( rc )
           return 2;

        anzahl++;
       }

     if ( ! Nutzerdef )
        Nutzerdateiname[0] = '\0';

     /* copy single files */
     for ( cnt = 0; cnt < anzahl; cnt++ )
       {
        if ( Fahrzeuge )
          {
           if ( fversion < ZOOMXARCHIVEFZGFORMAT_386_0 )
             {
              rc = ReadInt32( file1, &fzgnr );
              if ( rc )
                 return 2;
              fzgsatznr = fzgnr >> 4;
              fzgloknr = fzgnr & (FZGF_MAXFZGZAHL-1);
             }
           else /* fversion >= ZOOMXARCHIVEFZGFORMAT_386_0 */
             {
              rc = ReadInt32( file1, &fzgsatznr );
              if ( rc )
                 return 2;
              rc = ReadInt32( file1, &fzgloknr );
              if ( rc )
                 return 2;
              fzgnr = (fzgsatznr << 4) | fzgloknr;
             }
          }
        else /* Scenery/way */
          {
           rc = ReadInt32( file1, &symnr );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &bauform );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &fwsig );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &phase );
           if ( rc )
              return 2;
           rc = ReadInt32( file1, &alt );
           if ( rc )
              return 2;
          }

        rc = ReadInt32( file1, &len );
        if ( rc )
           return 2;

        rc = ReadInt32( file1, &Zeit );
        if ( rc )
           return 2;

        strcpy( PfadKpl, Pfad );

        if ( Fahrzeuge )
          {
           if ( Nutzerdef )
             {
              sprintf( Name, "%s-%02d-%02d.",
                             Nutzerdateiname, fzgsatznr, fzgloknr );
             }
           else /* Standard archive */
             {
              sprintf( Name, "%05d-%02d.", fzgsatznr, fzgloknr );
             }
              
           if ( Zoomfaktor == 2 )
              strcat( Name, FEXT_ZOOM2FZG );
           else
              strcat( Name, FEXT_ZOOM4FZG );
          }
        else
          {
           GetGzgDateiName( symnr, bauform, fwsig, alt, phase,
                            Name, Zoomfaktor, Nutzerdateiname );
          }

        strcat( PfadKpl, "\\" );
        strcat( PfadKpl, Name );

        printf( PfadKpl );
        printf( " %s%d", GetText( TXTNR_012 ), len );
        printfNL();

        if ( ! Ueberschreiben )
          {
           if ( DateiGibtEs( PfadKpl ) )
             {
              /* File already exists */
              printf( GetText( TXTNR_013 ) );
              printfNL();
              schreiben = FALSE;
             }
           else
              schreiben = TRUE;
          }
        else
           schreiben = TRUE;

        /* if ! Simulieren: however write = TRUE for counting/test */

        if (( schreiben ) && ( ! Simulieren ))
          {
           file2 = OpenWrite( PfadKpl );
           if ( file2 == NULL )
              return 4;
          }

        while ( len-- )
          {
           rc = ReadByte( file1, &b );
           if ( rc )
             {
              if (( schreiben ) && ( ! Simulieren ))
                 CloseWrite( file2 );
              return 2;
             }
           if (( schreiben ) && ( ! Simulieren ))
             {
              rc = WriteByte( file2, b );
              if ( rc )
                {
                 CloseRead( file1 );
                 CloseWrite( file2 );
                 return 4;
                }
             }   
          }

        if (( schreiben ) && ( ! Simulieren ))
          {
           rc = CloseWrite( file2 );
           if ( rc )
             {
              CloseRead( file1 );
              return 4;
             }

           /* Set date and time of the single file
              to do so, open again for writing
              (the last CloseWrite() may have written to the buffer,
               but after setftime() we should not write again) */
           if ( SetDateiAttr( PfadKpl, Zeit ) )
             {
              CloseRead( file1 );
              return 4;
             }

           anzahlkop++;
          } 
        else /* ! schreiben */
           anzahlueb++;
       }

     CloseRead( file1 );

     if ( pFound != NULL )
        *pFound = anzahl;
     if ( pCopied != NULL )
        *pCopied = anzahlkop;
     if ( pSkipped != NULL )
        *pSkipped = anzahlueb;

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL InfoUserEnd(VOID)
     /* Print Info text about parameters */
    {
     Meldung( GetText( TXTNR_006 ), GetText( TXTNR_005 ) );
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL Fehler( PCCHAR8 pText )
  {
     /* Print Error message */
     Meldung( pText, GetText( TXTNR_004 ) );
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL Trace( PCCHAR8 pText )
  {
     /* Print Trace message (for debug purposes only) */
     Meldung( pText, "Trace" );  /* language ENG always */
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL Meldung( PCCHAR8 pText, PCCHAR8 pTitel )
    {
     /* could be redirected to an output file or stdout/stderr... */
     MessageBox( NULL, pText, pTitel,
                 MB_OK | MB_ICONINFORMATION | MB_APPLMODAL );
    }

/* --------------------------------------------------------------------- */

STATIC CHAR FASTCALL Caps( CHAR zei )
     /* OUT: Grossbuchstabe wenn moegl. */
     {
      if (zei < 'a')
	 return zei;

      if (zei < ('z'+1))
	 return LOBYTE( zei - 32 );

      return zei;
     }

/*--------------------------------------------------------------------*/

VOID FASTCALL CapsString( PCHAR Text )
  /* wandelt Text in Grossbuchstaben um
     (auch mehrzeilig, Ende \0) */
  {
   INT32 i = 0;

     while ( Text[i] != '\0' )
       {
	Text[i] = Caps( Text[i] );
	i++;
       }
    }

/*--------------------------------------------------------------------*/

STATIC BOOL FASTCALL DateiGibtEs( PCCHAR8 pName )
   /* Prueft auf Vorhandensein einer Datei
      Checks whether file exists
      IN:   pName = Name der Datei, ggf. mit Pfad
      OUT:  TRUE: Datei existiert
            FALSE: nicht gefunden
            */
  {
   FILE * file;
   CHAR8 Buf[ MAXPATH ];
   INT32 len;

     if ( pName == NULL )
        return FALSE;
     len = strlen( pName );
     if (( len >= sizeof( Buf ) ) || ( len == 0 ))
        return FALSE;
     MemCopy1( Buf, pName, len+1 );

     file = fopen( Buf, "rb" );
     if ( file == NULL )
        return FALSE;

     fclose( file );
     return TRUE;
    }

/*--------------------------------------------------------------------*/

STATIC FILE * FASTCALL OpenWrite( PCCHAR8 pName )
   /* Oeffnet Datei zum Schreiben
      IN:   pName = Name der Datei
      OUT:  NULL Fehler (Datei nicht eroeffnet)
            sonst Handle
            */
  {
   FILE * file;

     file = fopen( pName, "wb" );
     if ( file == NULL )
        return NULL;

     WritePos = 0;
     return file;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL WriteByte( FILE * file2, BYTE b )
   /* schreibt Byte auf Ausgabedatei
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( WritePos < FILEBUFSIZE )
       {
        WriteBuf[ WritePos++ ] = b;
        return 0;
       }

     if ( fwrite( WriteBuf, FILEBUFSIZE, 1, file2 ) != 1 )
       {
        fclose( file2 );
        return 1;
       }

     WritePos = 1;
     WriteBuf[ 0 ] = b;

     return 0;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL WriteWord16( FILE * file2, WORD16 w )
   /* schreibt Word auf Ausgabedatei
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( WriteByte( file2, LOBYTE( w ) ) )
        return 1;

     return WriteByte( file2, HIBYTE( w ) );
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL WriteInt32( FILE * file2, INT32 w )
   /* schreibt INT32 auf Ausgabedatei
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler (Datei schon geschlossen)
            */
  {
     if ( WriteWord16( file2, LOWORD( w ) ) )
        return 1;

     return WriteWord16( file2, HIWORD( w ) );
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL CloseWrite( FILE * file2 )
   /* schliet Ausgabedatei, mit Flush des restlichen Puffers
      IN:   file2 = FileHandle
      OUT:  0 OK
            1 Fehler
            */
  {
   BOOL err = FALSE;

     if ( WritePos > 0 )
       {
        if ( fwrite( WriteBuf, WritePos, 1, file2 ) != 1 )
           err = TRUE;
       }

     if ( fclose( file2 ) != 0 )
        err = TRUE;

     if ( err )
        return 1;
     else
        return 0;
    }

/*--------------------------------------------------------------------*/

STATIC FILE * FASTCALL OpenRead( PCCHAR8 pName )
   /* Opens file for reading (buffered)
      IN:   pName = Name of file (may include path)
      OUT:  NULL: error (file not found / not opened)
            else: Handle
            */
  {
   FILE * file;

     file = fopen( pName, "rb" );
     if ( file == NULL )
        return NULL;

     ReadPos = FILEBUFSIZE;
     return file;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL ReadByte( FILE * file, PBYTE pB )
   /* reads BYTE from input file
      IN:   file = FileHandle
      OUT:  0 OK
            1 Error (file already closed)
            */
  {
     if ( ReadPos < FILEBUFSIZE )
       {
        *pB = ReadBuf[ ReadPos++ ];
        return 0;
       }

     if ( fread( ReadBuf, 1, FILEBUFSIZE, file ) == 0 )
        /* at ende of file maybe  rc < FILEBUFSIZE */
       {
        fclose( file );
        return 1;
       }

     ReadPos = 1;
     *pB = ReadBuf[ 0 ];

     return 0;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL ReadWord16( FILE * file, PWORD16 pW )
   /* reads WORD16 from input file
      IN:   file = FileHandle
      OUT:  0 OK
            1 Error (file already closed)
            */
  {
   BYTE bl;
   WORD16 w;

     if ( ReadByte( file, &bl ) )
        return 1;
     if ( ReadByte( file, (PBYTE) &w ) )
        return 1;

     *pW = LOWORD( ( w << 8 ) | bl );

     return 0;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL ReadInt32( FILE * file, PINT32 pI )
   /* reads INT32 from input file
      IN:   file = FileHandle
      OUT:  0 OK
            1 Error (file already closed)
            */
  {
   WORD16 wl, wh;

     if ( ReadWord16( file, &wl ) )
        return 1;
     if ( ReadWord16( file, &wh ) )
        return 1;

     *pI = ( wh << 16 ) | wl;

     return 0;
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL ReadVorspann( FILE * file )
   /* reads and skips ASCI header text from file
      IN:   file = FileHandle
      OUT:  0 OK
            1 Error (file already closed)
            */
  {
   BYTE  b;
   INT32 rc;

     do
       {
        rc = ReadByte( file, &b );
        if ( rc )
           return 2;
       }
     while ( b != CTRL_Z );

     return 0;
    } 

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL CloseRead( FILE * file )
   /* closes input file
      IN:   file = FileHandle
      OUT:  0 OK
            1 Error (for reading normally unimportant)
            */
  {
     if ( fclose( file ) != 0 )
        return 1;

     return 0;
    }

/* --------------------------------------------------------------------- */

STATIC PVOID FASTCALL AllocateMemory( INT32 len )
  /* allocates main memory
     IN:  len = Length in BYTE
    OUT:  NULL some error
          else pointer to the block */
  {
#ifndef USE_MEM_STANDARD_LIBRARY
   HGLOBAL hMem;
#endif   

     if ( len <= 0 )
        return NULL;

#ifdef USE_MEM_STANDARD_LIBRARY
     /* then set Linker Option "max. heap size" in a suitable way! */
     return malloc( len );
#else
     hMem = GlobalAlloc( GMEM_FIXED, len );
     if ( hMem == NULL )
        return NULL;
     return GlobalLock( hMem );
#endif
    }

/* --------------------------------------------------------------------- */

STATIC VOID FASTCALL FreeMemory( PVOID pDat )
  /* frees allocated memory block
     IN:  pDat = Block */
    {
#ifndef USE_MEM_STANDARD_LIBRARY
   HGLOBAL hMem;
#endif

     if ( pDat == NULL )
        return;

#ifdef USE_MEM_STANDARD_LIBRARY
     free( pDat );
#else
     hMem = GlobalHandle( pDat );
     if ( hMem == NULL )
        return;

     GlobalUnlock( hMem );
     GlobalFree( hMem );
#endif
    }

/*--------------------------------------------------------------------*/

STATIC INT32 FASTCALL ReadOverSpacePos( PCCHAR8 pBuf )
 /* ueberliest Leerzeichen in Puffer pBuf
    IN:  pBuf = beliebiger Pufferbereich
    OUT: Offset erstes Nicht-' '
         (kann 0, d.h.  gleich der Eingabeposition sein) */
  {
   INT32 Pos = 0;

     while ( pBuf[ Pos ] == ' ' )
        Pos++;
     return Pos;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL DelSpaces( PCHAR8 pBuf )
 /* eliminiert Leerzeichen am Anfang des Puffers pBuf
    IN:  pBuf = beliebiger Pufferbereich  (Ende \0)
    OUT: pBuf = dasselbe, fuehrende Leerzeichen weggefallen
           rc = Anzahl geloeschte Leerzeichen */
  {
   INT32 Pos = ReadOverSpacePos( pBuf );
   INT32 Len;

     if ( Pos > 0 )
       {
        Len = strlen( pBuf + Pos );
        MemMove1( pBuf, pBuf + Pos, Len + 1 ); /* inkl. \0 */
       }

     return Pos;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL SetDateiAttr( PCCHAR8 pPfad, INT32 Zeit )
  /* setzt Datum und Uhrzeit
     IN:  pPfad = pPfad+Datei
          Zeit = HIWORD = Datum, LOWORD = Uhrzeit
                 im Format von findfirst()/findnext()
     OUT: 0 OK
          1 Datei nicht gefunden
          2 Schreibfehler  */
  {
   INT32 hfile, rc, rcc, datum, uhrzeit;
   struct ftime ftime;

     /* Datum + Zeit setzen
        dazu nochmal zum schreiben oeffnen
        (das letzte CloseWrite hat ggf. noch Puffer geschrieben, nach
         setftime darf aber kein schreiben mehr erfolgen) */
     hfile = open( pPfad, O_WRONLY | O_BINARY );
     if ( hfile == -1 )
        return 1;

     /* Bit0..4 Tag, Bit5..8 Monat, Bit9..15 Jahr seit 1980 */
     datum = Zeit >> 16;

     /* Bit0..4 sec/2, Bit5..10 min, Bit11..15 h */
     uhrzeit = LOWORD( Zeit );

     ftime.ft_tsec = uhrzeit & 0x001F;
     ftime.ft_min  = ( uhrzeit >> 5 ) & 0x003F;
     ftime.ft_hour = ( uhrzeit >> 11 ) & 0x001F;

     ftime.ft_day   = datum & 0x001F;
     ftime.ft_month = ( datum >> 5 ) & 0x000F;
     ftime.ft_year  = ( datum >> 9 ) & 0x007F;

     rc = setftime( hfile, &ftime );
     rcc = close( hfile );

     if (( rc == -1 ) || ( rcc == -1 ))
        return 2;
     else
        return 0;
    }

/* --------------------------------------------------------------------- */

STATIC INT32 FASTCALL GetSprache( VOID )
   /* Sprache ermitteln / Get Language
      OUT: SPRACHE_XXX */
  {
   LANGID sprachID;
   INT32  spcode, sprachIDPrim;

     /* get user language (Windows) */
     sprachID = GetUserDefaultLangID();

     /* primary ID is enough here: there are many sub-ID... */
     sprachIDPrim = PRIMARYLANGID( sprachID );
     switch( sprachIDPrim )
       {
        case LANG_GERMAN:
           spcode = SPRACHE_DEU;
           break;

        case LANG_SPANISH:
           spcode = SPRACHE_ESP;
           break;

        case LANG_CZECH:
        case LANG_SLOVAK: /* also Czech because similar */
           spcode = SPRACHE_CZE;
           break;

        case LANG_POLISH:
           spcode = SPRACHE_POL;
           break;

        default: /* including LANG_ENGLISH */
           spcode = SPRACHE_ENG;
       }

     return spcode;
    }

/* --------------------------------------------------------------------- */

STATIC PCCHAR8 FASTCALL GetText( INT32 Textnummer )
   /* Get text (depending on Sprach_Code)
      IN:  Textnummer  = Text number
           Sprach_Code = current language
      OUT: pointer to the text */
  {
   STATIC  CHAR8 errtext[100]; /* STATIC for giving back! */
   INT32   spcode, pos;
   PCCHAR8 pTxt;

/* Polskie tumaczenie / Polnische Uebersetzung: . M. P. P. 12.2008
   Czesky tumaczenie / Polnische Uebersetzung: . M. P. P. 12.2008
    (and P.Macku 1/2009)
   Some additions made by JB using Google Translation Services 
   Remark JB about special characters:
    Correct coding Win/ANSI or DOS depends on purpose:
     Texts for console output: DOS Coding (here marked as "stdout")
           for messagebox output: Win/ANSI
     However: may be system dependent */

   STATIC CHAR8 txt00de[] = "Text fehlt:";
   STATIC CHAR8 txt00en[] = "Text missing:";
   STATIC CHAR8 txt00pl[] = "Brakuje tekstu:";
   STATIC CHAR8 txt00es[] = "Texto falta:";
   STATIC CHAR8 txt00cz[] = "Chyb text:";

   /* stdout */
   STATIC CHAR8 txt01de[] = "Version";
   STATIC CHAR8 txt01en[] = "Version";
   STATIC CHAR8 txt01pl[] = "Wersja";
   STATIC CHAR8 txt01es[] = "versin";
   STATIC CHAR8 txt01cz[] = "Verze";

   STATIC CHAR8 txt02de[] = "Archivdatei \"%s\" nicht erffnet";
   STATIC CHAR8 txt02en[] = "Archive file \"%s\" not opened";
   STATIC CHAR8 txt02pl[] = "Plik archiwum %s nie zosta otwarty";
   STATIC CHAR8 txt02es[] = "No se puede abrir el fichero de archivo \"%s\"";
   STATIC CHAR8 txt02cz[] = "Archivn soubor \"%s\" nebyl oteven";

   STATIC CHAR8 txt03de[] = "Archivdatei \"%s\" existiert schon";
   STATIC CHAR8 txt03en[] = "Archive file \"%s\" already exists";
   STATIC CHAR8 txt03pl[] = "Plik archiwum %s ju istnieje";
   STATIC CHAR8 txt03es[] = "El fichero \"%s\" ya existe";
   STATIC CHAR8 txt03cz[] = "Archivn soubor \"%s\" ji existuje";

   STATIC CHAR8 txt04de[] = "Fehler";
   STATIC CHAR8 txt04en[] = "Error";
   STATIC CHAR8 txt04pl[] = "Bd";
   STATIC CHAR8 txt04es[] = "Error";
   STATIC CHAR8 txt04cz[] = "Chyba";

   STATIC CHAR8 txt05de[] = "Korrekter Aufruf:";
   STATIC CHAR8 txt05en[] = "Correct call:";
   STATIC CHAR8 txt05pl[] = "Prawidowe wywoanie:";
   STATIC CHAR8 txt05es[] = "Usando correcto:";
   STATIC CHAR8 txt05cz[] = "Sprvn syntaxe:";

   STATIC CHAR8 txt06de[] =
"FZGC2 Kommando Verzeichnis Dateiname [Optionen]\n\n\
Kommandos: V2|V4|W2|W4|S1|S2|S4|U1|U2|U4|X\n\n\
Vn: \"Verzeichnis\\?????-??.fzn\" => \"Dateiname.fzn\"\n\
Wn: \"Verzeichnis\\Dateiname-??-??.fzn\" => \"Dateiname.fzn\"\n\
Sn: \"Verzeichnis\\*.gzn\" => \"Dateiname\"\n\
Un: \"Verzeichnis\\u??-???-Dateiname.gzn\" => \"Dateiname[.uzn]\"\n\
X: in \"Dateiname\" enthaltene Dateien => \"Verzeichnis\\*.*\"\n\n\
Optionen: [O][S][I]\n\n\
O: berschreiben existierender Dateien\n\
S: Simulationsmodus (nichts schreiben)\n\
I: Informationen/Statistik ausgeben\n\n\
Keine Leerzeichen in Namen\n\
Kommandos/Optionen Gro-/Kleinschreibung egal";
   STATIC CHAR8 txt06en[] =
"FZGC2 command directory filename [options]\n\n\
Commands: V2|V4|W2|W4|S1|S2|S4|U1|U2|U4|X\n\
Vn: \"directory\\?????-??.fzn\" => \"filename.fzn\"\n\
Wn: \"directory\\filename-??-??.fzn\" => \"filename.fzn\"\n\
Sn: \"directory\\*.gzn\" => \"filename\"\n\
Un: \"directory\\u??-???-filename.gzn\" => \"filename[.uzn]\"\n\
X: in \"filename\" stored files => \"directory\\*.*\"\n\n\
Options: [O][S][I]\n\
O: Overwrite existing files\n\
S: Simulation mode (write nothing)\n\
I: Informations/Statistics\n\n\
No spaces in names\n\
Commands/Options case insensitive";
   STATIC CHAR8 txt06pl[] =
"FZGC2 polecenie cieka nazwa_pliku [opcje]\n\n\
Polecenia: V2|V4|W2|W4|S1|S2|S4|U1|U2|U4|X\n\
Vn: cieka\\?????-??.fzn => nazwa_pliku.fzn\n\
Wn: cieka\\nazwa_pliku-??-??.fzn => nazwa_pliku.fzn\n\
Sn: cieka\\*.gzn => nazwa_pliku\n\
Un: cieka\\u??-???-nazwa_pliku.gzn => nazwa_pliku[.uzn]\n\
X: pliki zawarte w nazwa_pliku => cieka\\*.*\n\n\
Opcje: [O][S][I]\n\
O: Nadpisuj istniejce pliki\n\
S: Tryb symulacji (bez zapisu)\n\
I: Informacje/Statystyka\n\n\
Nazwy plikw/katalogw nie mog zawiera spacji\n\
Wielko liter nie jest uwzgldniana w poleceniach/opcjach";
   STATIC CHAR8 txt06es[] =
"FZGC2 comando directorio nombre_de_fichero [opciones]\n\n\
Comandos: V2|V4|W2|W4|S1|S2|S4|U1|U2|U4|X\n\
Vn: \"directorio\\?????-??.fzn\" => \"nombre.fzn\"\n\
Wn: \"directorio\\nombre-??-??.fzn\" => \"nombre.fzn\"\n\
Sn: \"directorio\\*.gzn\" => \"nombre\"\n\
Un: \"directorio\\u??-???-nombre.gzn\" => \"nombre[.uzn]\"\n\
X: ficheros guardados en \"nombre\" => \"directorio\\*.*\"\n\n\
Opciones: [O][S][I]\n\
O: reemplazar ficheros existente\n\
S: modo de la simulacin (no escribir nada)\n\
I: informacin/estadstica\n\n\
No espacios en nombres\n\
Comandos/opciones insignificantes si con mayscula/minscula";
   STATIC CHAR8 txt06cz[] =
"FZGC2 pkaz cesta nzev_souboru [volby]\n\n\
Pkazy: V2|V4|W2|W4|S1|S2|S4|U1|U2|U4|X\n\
Vn: \"cesta\\?????-??.FZn\" => \"nzev_souboru.FZn\"\n\
Wn: \"cesta\\nzev_souboru-??-??.FZn\" => \"nzev_souboru.FZn\"\n\
Sn: \"cesta\\*.GZn\" => \"nzev_souboru\"\n\
Un: \"cesta\\u??-???-nzev_souboru.GZn\" => \"nzev_souboru[.UZn]\"\n\
X: soubory obsaen v \"nzev_souboru\" => \"cesta\\*.*\"\n\n\
Volby: [O][S][I]\n\
O: Pepsat existujc soubory\n\
S: Md simulace (bez zpisu)\n\
I: Informace/Statistika\n\n\
Nzvy soubor a sloek nesm obsahovat mezery\n\
Pkazy a volby nerozliuj velikost psmen.";

   STATIC CHAR8 txt07de[] =
"Dateien gefunden: %d  ausgepackt: %d  bersprungen: %d";
   STATIC CHAR8 txt07en[] =
"Files found: %d  extracted: %d  skipped: %d";
   STATIC CHAR8 txt07pl[] =
"Znaleziono plikw: %d  wypakowano: %d  pominito: %d";
   STATIC CHAR8 txt07es[] =
"Ficheros encontrados: %d  generados: %d  cancelados: %d";
   STATIC CHAR8 txt07cz[] =
"Nalezeno soubor: %d  Rozbaleno: %i  Peskoeno: %d";

   /* multiple used! */
   STATIC CHAR8 txt08de[] = "Ende/fertig";
   STATIC CHAR8 txt08en[] = "Done/ready";
   STATIC CHAR8 txt08pl[] = "Gotowe";
   STATIC CHAR8 txt08es[] = "Terminado";
   STATIC CHAR8 txt08cz[] = "Hotovo";

   STATIC CHAR8 txt09de[] = "Auspacken: rc=";
   STATIC CHAR8 txt09en[] = "Extracting: rc=";
   STATIC CHAR8 txt09pl[] = "Wypakowywanie: kod zwrotny=";
   STATIC CHAR8 txt09es[] = "Descomprimir: rc=";
   STATIC CHAR8 txt09cz[] = "Rozbalovn: rc=";

   STATIC CHAR8 txt10de[] = "Eingepackt: %i Dateien";
   STATIC CHAR8 txt10en[] = "Packed: %i files";
   STATIC CHAR8 txt10pl[] = "Spakowano: %i plikw";
   STATIC CHAR8 txt10es[] = "Comprimiendos: %i ficheros";
   STATIC CHAR8 txt10cz[] = "Sbaleno: %i soubor";

   STATIC CHAR8 txt11de[] = "Einpacken: rc=";
   STATIC CHAR8 txt11en[] = "Packing: rc=";
   STATIC CHAR8 txt11pl[] = "Pakowanie: kod zwrotny=";
   STATIC CHAR8 txt11es[] = "Comprimir: rc=";
   STATIC CHAR8 txt11cz[] = "Balen: rc=";

   /* stdout */
   STATIC CHAR8 txt12de[] = "Laenge=";
   STATIC CHAR8 txt12en[] = "Length=";
   STATIC CHAR8 txt12pl[] = "Dugo=";
   STATIC CHAR8 txt12es[] = "La longitud=";
   STATIC CHAR8 txt12cz[] = "Dlka="; /* Delka */

   /* stdout */
   STATIC CHAR8 txt13de[] = "Datei existiert schon";
   STATIC CHAR8 txt13en[] = "File already exists";
   STATIC CHAR8 txt13pl[] = "Plik ju istnieje";
   STATIC CHAR8 txt13es[] = "El fichero ya existe";
   STATIC CHAR8 txt13cz[] = "Soubor ji existuje"; /* S. jiz e. */

   STATIC CHAR8 txt14de[] = "Keine Datei gefunden";
   STATIC CHAR8 txt14en[] = "No file found";
   STATIC CHAR8 txt14pl[] = "Plik nie zosta znaleziony";
   STATIC CHAR8 txt14es[] = "Ninguno fichero no encuentro";
   STATIC CHAR8 txt14cz[] = "Soubor nenalezen";

   STATIC CHAR8 txt15de[] = "Speicher reicht nicht";
   STATIC CHAR8 txt15en[] = "Out of memory";
   STATIC CHAR8 txt15pl[] = "Zabrako pamici";
   STATIC CHAR8 txt15es[] = "No hay memoria";
   STATIC CHAR8 txt15cz[] = "Nedostatek pamti";

   STATIC CHAR8 txt16de[] = "Keine Archivdatei erzeugt";
   STATIC CHAR8 txt16en[] = "No archive file created";
   STATIC CHAR8 txt16pl[] = "Plik archiwum nie zosta utworzony";
   STATIC CHAR8 txt16es[] = "Ninguno fichero de archivo no generado";
   STATIC CHAR8 txt16cz[] = "Archivn soubor nebyl vytvoen";

   STATIC CHAR8 txt17de[] = "Schreibfehler";
   STATIC CHAR8 txt17en[] = "Error while writing";
   STATIC CHAR8 txt17pl[] = "Bd podczas zapisu";
   STATIC CHAR8 txt17es[] = "Error al escribir";
   STATIC CHAR8 txt17cz[] = "Chyba bhem zpisu";

   STATIC CHAR8 txt18de[] = "Lesefehler";
   STATIC CHAR8 txt18en[] = "Error while reading";
   STATIC CHAR8 txt18pl[] = "Bd odczytu";
   STATIC CHAR8 txt18es[] = "Error de lectura";
   STATIC CHAR8 txt18cz[] = "Chyba ten souboru";

   STATIC CHAR8 txt19de[] = "Datenfehler";
   STATIC CHAR8 txt19en[] = "Data error";
   STATIC CHAR8 txt19pl[] = "Bd danych";
   STATIC CHAR8 txt19es[] = "Error de datos";
   STATIC CHAR8 txt19cz[] = "Chyba ten dat";

   STATIC CHAR8 txt20de[] = "Dateien sammeln:";
   STATIC CHAR8 txt20en[] = "Collecting files";
   STATIC CHAR8 txt20pl[] = "Zbieranie plikw";
   STATIC CHAR8 txt20es[] = "Recopilacin de los archivos";
   STATIC CHAR8 txt20cz[] = "Sber souboru";

   STATIC CHAR8 txt21de[] = "Dateien kopieren:";
   STATIC CHAR8 txt21en[] = "Copying files";
   STATIC CHAR8 txt21pl[] = "Kopiowanie plikw";
   STATIC CHAR8 txt21es[] = "Copia de archivos";
   STATIC CHAR8 txt21cz[] = "Koprovn souboru";

   STATIC CHAR8 txt22de[] = "Front2";
   STATIC CHAR8 txt22en[] = "front2";
   STATIC CHAR8 txt22pl[] = "front2";
   STATIC CHAR8 txt22es[] = "frente2";
   STATIC CHAR8 txt22cz[] = "predn2";

   STATIC CHAR8 txt23de[] = "Seite2";
   STATIC CHAR8 txt23en[] = "side2";
   STATIC CHAR8 txt23pl[] = "strona2";
   STATIC CHAR8 txt23es[] = "lado2";
   STATIC CHAR8 txt23cz[] = "strana2";

   /* "rl" = "right left" */
   STATIC CHAR8 txt24de[] = "Tr2,r,l";
   STATIC CHAR8 txt24en[] = "door2,r,l";
   STATIC CHAR8 txt24pl[] = "drzwi2,p,z";
   STATIC CHAR8 txt24es[] = "puerta2,d,i";
   STATIC CHAR8 txt24cz[] = "dvere2,v,l";

   /* length */
   STATIC CHAR8 txt25de[] = "l";
   STATIC CHAR8 txt25en[] = "l";
   STATIC CHAR8 txt25pl[] = "d";
   STATIC CHAR8 txt25es[] = "l";
   STATIC CHAR8 txt25cz[] = "d";

   /* height, side view (horizontal/waagerecht) */
   STATIC CHAR8 txt26de[] = "hs";
   STATIC CHAR8 txt26en[] = "hs";
   STATIC CHAR8 txt26pl[] = "ws";
   STATIC CHAR8 txt26es[] = "ac";
   STATIC CHAR8 txt26cz[] = "vb";

   /* width, front/rear view */
   STATIC CHAR8 txt27de[] = "b";
   STATIC CHAR8 txt27en[] = "w";
   STATIC CHAR8 txt27pl[] = "s";
   STATIC CHAR8 txt27es[] = "a";
   STATIC CHAR8 txt27cz[] = "s";

   /* height, front/rear view (vertically/senkrecht) */
   STATIC CHAR8 txt28de[] = "hf";
   STATIC CHAR8 txt28en[] = "hf";
   STATIC CHAR8 txt28pl[] = "wf";
   STATIC CHAR8 txt28es[] = "af";
   STATIC CHAR8 txt28cz[] = "vp";

   /* front/rear view 2nd,   side view 2nd,   stop view 2nd (open doors) */
   STATIC CHAR8 txt29de[] = " f=%c s=%c t=%c";
   STATIC CHAR8 txt29en[] = " f=%c s=%c d=%c";
   STATIC CHAR8 txt29pl[] = " f=%c s=%c d=%c";
   STATIC CHAR8 txt29es[] = " f=%c l=%c p=%c";
   STATIC CHAR8 txt29cz[] = " p=%c s=%c d=%c";

   /* [0] front/rear view 2nd = no
      [1] front/rear view 2nd = yes
      [2] side view 2nd = no
      [3] side view 2nd = yes
      [4] stop view 2nd (open doors) = no
      [5] stop view 2nd (open doors) = right side only
      [6] stop view 2nd (open doors) = left side only
      [7] stop view 2nd (open doors) = both sides
      */
   STATIC CHAR8 txt30de[] = "1212-RL2";
   STATIC CHAR8 txt30en[] = "1212-RL2";
   STATIC CHAR8 txt30pl[] = "1212-PZ2";
   STATIC CHAR8 txt30es[] = "1212-di2";
   STATIC CHAR8 txt30cz[] = "1212-VL2";

   CONST TXTELEM pTxte[] = {
    /* Order like Text numbers (but not needed) and SPRACHE_xxx (needed)
       End marker = TXTNR_UNDEF */
     { TXTNR_000, { txt00en, txt00de, txt00pl, txt00es, txt00cz }},
     { TXTNR_001, { txt01en, txt01de, txt01pl, txt01es, txt01cz }},
     { TXTNR_002, { txt02en, txt02de, txt02pl, txt02es, txt02cz }},
     { TXTNR_003, { txt03en, txt03de, txt03pl, txt03es, txt03cz }},
     { TXTNR_004, { txt04en, txt04de, txt04pl, txt04es, txt04cz }},
     { TXTNR_005, { txt05en, txt05de, txt05pl, txt05es, txt05cz }},
     { TXTNR_006, { txt06en, txt06de, txt06pl, txt06es, txt06cz }},
     { TXTNR_007, { txt07en, txt07de, txt07pl, txt07es, txt07cz }},
     { TXTNR_008, { txt08en, txt08de, txt08pl, txt08es, txt08cz }},
     { TXTNR_009, { txt09en, txt09de, txt09pl, txt09es, txt09cz }},
     { TXTNR_010, { txt10en, txt10de, txt10pl, txt10es, txt10cz }},
     { TXTNR_011, { txt11en, txt11de, txt11pl, txt11es, txt11cz }},
     { TXTNR_012, { txt12en, txt12de, txt12pl, txt12es, txt12cz }},
     { TXTNR_013, { txt13en, txt13de, txt13pl, txt13es, txt13cz }},
     { TXTNR_014, { txt14en, txt14de, txt14pl, txt14es, txt14cz }},
     { TXTNR_015, { txt15en, txt15de, txt15pl, txt15es, txt15cz }},
     { TXTNR_016, { txt16en, txt16de, txt16pl, txt16es, txt16cz }},
     { TXTNR_017, { txt17en, txt17de, txt17pl, txt17es, txt17cz }},
     { TXTNR_F_LESEN, { txt18en, txt18de, txt18pl, txt18es, txt18cz }},
     { TXTNR_F_DATEN, { txt19en, txt19de, txt19pl, txt19es, txt19cz }},
     { TXTNR_COLLECT, { txt20en, txt20de, txt20pl, txt20es, txt20cz }},
     { TXTNR_COPY,    { txt21en, txt21de, txt21pl, txt21es, txt21cz }},
     { TXTNR_FRONT_2, { txt22en, txt22de, txt22pl, txt22es, txt22cz }},
     { TXTNR_SEITE_2, { txt23en, txt23de, txt23pl, txt23es, txt23cz }},
     { TXTNR_HST_2RL, { txt24en, txt24de, txt24pl, txt24es, txt24cz }},
     { TXTNR_LEN,     { txt25en, txt25de, txt25pl, txt25es, txt25cz }},
     { TXTNR_HOS,     { txt26en, txt26de, txt26pl, txt26es, txt26cz }},
     { TXTNR_BRF,     { txt27en, txt27de, txt27pl, txt27es, txt27cz }},
     { TXTNR_HOF,     { txt28en, txt28de, txt28pl, txt28es, txt28cz }},
     { TXTNR_F_S_T,   { txt29en, txt29de, txt29pl, txt29es, txt29cz }},
     { TXTNR_F_S_T_Z, { txt30en, txt30de, txt30pl, txt30es, txt30cz }},
#if ( SPRACHE_ANZAHL > 5 )
#error Add some NULLs!
#endif      
     { TXTNR_UNDEF, { NULL, NULL, NULL, NULL, NULL }}
     };

    if (( Sprach_Code < SPRACHE_MIN ) || ( Sprach_Code > SPRACHE_MAX ))
       spcode = SPRACHE_ENG;
    else
       spcode = Sprach_Code;

    pos = 0;
    while ( pTxte[pos].TxtNr != TXTNR_UNDEF )
      {
       if ( pTxte[pos].TxtNr == Textnummer )
          break;
       pos++;
      }

    pTxt = pTxte[pos].pTxt[spcode];
    if ( pTxt == NULL )
      {
       sprintf( errtext, "%s %i",
                pTxte[0].pTxt[spcode], Textnummer );
       return errtext;
      }

    return pTxt;
   }

/*--------------------------------------------------------------------*/

STATIC VOID FASTCALL MemMove1( PVOID pDest, PCVOID pSrc, INT32 len )
/* same as memmove()
   copies byte[len] from pSrc to pDest
   with correct handling of overlapping areas */
    {
     if (( pDest == NULL ) || ( pSrc == NULL ) || ( len <= 0 ))
        return;

#ifdef TEST310
     if ( len >= 32767 )
       {
        CHAR8 bbb[80];
        printf( bbb, "MemMove1 len = %u ", len );
        Trace( bbb );
        return;
       }
     if ( pDest == pSrc )
       {
        CHAR8 bbb[80];
        printf( bbb, "MemMove1 Dest==Src " );
        Trace( bbb );
        return;
       }
#endif

     memmove( pDest, pSrc, len );
    }

/*------------------------------------------------------- */

STATIC INT32 FASTCALL GetGzgDateiName( INT32 GSym, INT32 Bauform, INT32 FwSig,
                                       INT32 Alt, INT32 Phase,
                                       PCHAR8 pFName, INT32 Zf,
                                       PCCHAR8 pNutzerName )
/* gets filename for a single graphics file for a scenery/way/signal symbol
  IN: GSym = Symbol number (scenery/userdefined/way)
      Bauform = for way: Design (version)
      FwSig   = for way: HIWORD: driving way info
                         LOWORD: for signals phase 1/2/3 (a/b/c)
      Alt = for scenery: KEINE_ALTERNATIVE = there is only this version
                         MIN_ALTERNATIVE..MAX_ = exactly this is meant
      Phase   = Phase number if animation (0 if no)
      pFName = buffer for receiving the name [MAXPATH]
      Zf = Zoom factor (1/2/4)
      pNutzerName = Name of user-defined file or empty for standard
 OUT: 0 OK, Standard symbol (use resp. standard path or std. archive file)
      1 OK, User-defined symbol (use data path)
      2 wrong parameters (symbol umber/zoom factor)
	    */
  {
   INT32   AddZeichen, FwZei;
   CHAR8   ElemBuf[40], PhaseBuf[20], BauformBuf[20];
   PCHAR8  pExt;
   BOOL    nutzerdef, fahrweg;

     if ( pFName == NULL )
        return 2;

     if ( pNutzerName[0] != '\0' )
        /* user-defined */
       {
        nutzerdef = TRUE;
        /* since b386b2 order: "Name-nn-nnx" */
        strcpy( pFName, pNutzerName );
        stradd( pFName, '-' );
       }
     else
       {
        nutzerdef = FALSE;
        pFName[0] = '\0';
       }

     if (( GSym <= MAX_FAHRWEG ) && ( ! nutzerdef ))
        fahrweg = TRUE;
     else
        fahrweg = FALSE;

     if ( fahrweg )
       {
        Alt   = KEINE_ALTERNATIVE;
        Phase = -1; /* MIN_ANIPHASE; */
       }
     else
       {
        Bauform = 0;
       }

     if ( Zf == 1 )
        pExt = FEXT_ZOOM1GZG;
     else
        if ( Zf == 2 )
           pExt = FEXT_ZOOM2GZG;
        else
           if ( Zf == 4 )
              pExt = FEXT_ZOOM4GZG;
           else
              return 2;

     if (( Alt >= MIN_ALTERNATIVE ) && ( Alt <= MAX_ALTERNATIVE ))
        AddZeichen = 'a'-1 + Alt;  /* 1..4 => a..d */
     else
        AddZeichen = '\0';

     /* Since BAHN 3.86b2: also "-00" for MIN_ANIPHASE
        but not for std. driving way: bauform used instead */
     if ( Phase >= MIN_ANIPHASE )
        sprintf( PhaseBuf, "-%02d", Phase );
     else
        PhaseBuf[0] = '\0';

     /* Design by syntax as phase, but this one only for way,
        the other one for scenery only */
     if ( fahrweg )
       {
        if ( LOWORD( FwSig ) != 0 )
           /* Signal 'a'..'d' */
          {
           AddZeichen = 'a' + LOWORD( FwSig ) - 1;
          }

        sprintf( BauformBuf, "-%03d", Bauform );

        FwZei = Fahrwegzeichen[ HIWORD( FwSig )-1 ];

        /* create filename from symbol number
           (undependent on driving way) */
        sprintf( ElemBuf, "%c%03d", FwZei, GSym );

       }
     else /* no std. driving way */
       {
        BauformBuf[0] = '\0';

        if ( nutzerdef )
           /* user-defined */
          {
           /* elements 0..71 */
           sprintf( ElemBuf, "%02d", GSym );
          }
        else /* Standard */
          {
           /* create filename from symbol number */
           sprintf( ElemBuf, "%06d", GSym );
          }
       }

     strcat( pFName, ElemBuf );

     if ( PhaseBuf[0] != '\0' )
        strcat( pFName, PhaseBuf );

     if ( BauformBuf[0] != '\0' )
        strcat( pFName, BauformBuf );

     if ( AddZeichen != '\0' )
        stradd( pFName, LOBYTE( AddZeichen ));
     else /* since B386b2 always 'a' */
        stradd( pFName, 'a' );

     stradd( pFName, '.' );
     strcat( pFName, pExt );

     if ( nutzerdef )
        return 1;
     else
        return 0;
    }

/*------------------------------------------------------- */

STATIC VOID FASTCALL stradd( PCHAR pStr, CHAR Zei )
    /* appends a character to a string
       IN: pStr = String
           Zei  = Character */
  {
   CHAR str[2];

     str[0] = Zei;
     str[1] = '\0';
     strcat( pStr, str );
    }

/*------------------------------------------------------- */

STATIC VOID FASTCALL SubExt( PCHAR pName )
 /* removes extension if existing
    IN: pName = Buffer containing a filename
        The buffer may contain a path including '.' characters */
  {
   INT32  p = strlen( pName ) - 1;

     while (( p > 0 ) && ( pName[p] != '\\' ))
       {
        if ( pName[p] == L'.' )
          {
           pName[p] = '\0';
           return;
          }
        else
           p--;
       }
    }

/*------------------------------------------------------- */

STATIC VOID FASTCALL printfNL( VOID )
  /* output of CR LF (new line) to stdout */
    {

  /* printf( "\r\n" ); */

     printf( "\n" );
    }

